find_people_response = {
    "vanId": 19722445,
    "firstName": "Bob",
    "lastName": "Smith",
    "middleName": None,
    "suffix": None,
    "title": None,
    "contactMode": None,
    "organizationContactCommonName": None,
    "organizationContactOfficialName": None,
    "salutation": "Bob",
    "formalSalutation": "Bob Smith",
    "additionalSalutation": None,
    "preferredPronoun": None,
    "envelopeName": "Bob Smith",
    "formalEnvelopeName": "Bob Smith",
    "additionalEnvelopeName": None,
    "contactMethodPreferenceCode": None,
    "nickname": None,
    "website": None,
    "professionalSuffix": None,
    "party": None,
    "employer": None,
    "occupation": None,
    "sex": "F",
    "dateOfBirth": "1975-09-18T00:00:00Z",
    "selfReportedRace": None,
    "selfReportedEthnicity": None,
    "selfReportedRaces": None,
    "selfReportedEthnicities": None,
    "selfReportedGenders": None,
    "selfReportedSexualOrientations": None,
    "selfReportedLanguagePreference": None,
    "emails": None,
    "phones": None,
    "addresses": None,
    "recordedAddresses": None,
    "identifiers": None,
    "codes": None,
    "customFields": None,
    "contributionSummary": None,
    "suppressions": None,
    "caseworkCases": None,
    "caseworkIssues": None,
    "caseworkStories": None,
    "notes": None,
    "scores": None,
    "customProperties": None,
    "electionRecords": None,
    "membershipStatus": None,
    "organizationRoles": None,
    "districts": None,
}

get_person_response = {
    "vanId": 19722445,
    "firstName": "Bob",
    "lastName": "Smith",
    "middleName": None,
    "suffix": None,
    "title": None,
    "contactMode": None,
    "organizationContactCommonName": None,
    "organizationContactOfficialName": None,
    "salutation": "Bob",
    "formalSalutation": "Bob Smith",
    "additionalSalutation": None,
    "preferredPronoun": None,
    "envelopeName": "Bob Smith",
    "formalEnvelopeName": "Bob Smith",
    "additionalEnvelopeName": None,
    "contactMethodPreferenceCode": None,
    "nickname": None,
    "website": None,
    "professionalSuffix": None,
    "party": None,
    "employer": None,
    "occupation": None,
    "sex": "F",
    "dateOfBirth": "1975-09-18T00:00:00Z",
    "selfReportedRace": None,
    "selfReportedEthnicity": None,
    "selfReportedRaces": None,
    "selfReportedEthnicities": None,
    "selfReportedGenders": None,
    "selfReportedSexualOrientations": None,
    "selfReportedLanguagePreference": None,
    "emails": [],
    "phones": [
        {
            "phoneId": 15406767,
            "phoneNumber": "4142020792",
            "ext": None,
            "dateCreated": "2019-01-11T12:19:00Z",
            "phoneType": "Cell",
            "isPreferred": True,
            "smsOptInStatus": "Unknown",
            "phoneOptInStatus": "Unknown",
            "isCellStatus": {"statusId": 1, "statusName": "Verified Cell"},
        }
    ],
    "addresses": [
        {
            "addressId": None,
            "addressLine1": "900 N Washtenaw Ave",
            "addressLine2": None,
            "addressLine3": None,
            "city": "Chicago",
            "stateOrProvince": "IL",
            "zipOrPostalCode": "60622-4455",
            "geoLocation": None,
            "countryCode": "US",
            "preview": "900 N Washtenaw Ave ",
            "type": "Voting",
            "isPreferred": True,
            "streetAddress": "900 N Washtenaw Ave",
            "displayMode": "Standardized",
        },
        {
            "addressId": None,
            "addressLine1": "900 N Washtenaw Ave",
            "addressLine2": None,
            "addressLine3": None,
            "city": "Chicago",
            "stateOrProvince": "IL",
            "zipOrPostalCode": "60622-4455",
            "geoLocation": None,
            "countryCode": "US",
            "preview": "900 N Washtenaw Ave ",
            "type": "Mailing",
            "isPreferred": None,
            "streetAddress": "900 N Washtenaw Ave",
            "displayMode": "Standardized",
        },
    ],
    "recordedAddresses": [],
    "identifiers": [{"type": "StateFileID", "externalId": "KLYW682Z"}],
    "codes": None,
    "customFields": [],
    "contributionSummary": None,
    "suppressions": [],
    "caseworkCases": None,
    "caseworkIssues": None,
    "caseworkStories": None,
    "notes": [{"id": 1, "isViewRestricted": True, "text": "a"}],
    "scores": None,
    "customProperties": None,
    "electionRecords": None,
    "membershipStatus": None,
    "organizationRoles": None,
    "districts": None,
}

merge_contacts_response = {"vanId": 56789}

delete_person_response = {}



from parsons import Postgres, DBSync, Table, Redshift
from test.test_databases.fakes import FakeDatabase
from test.utils import assert_matching_tables
import unittest
import os

_dir = os.path.dirname(__file__)

TEMP_SCHEMA = "parsons_test"


# These tests interact directly with the Postgres database. In order to run, set the
# env to LIVE_TEST='TRUE'.
@unittest.skipIf(
    not os.environ.get("LIVE_TEST"), "Skipping because not running live test"
)
class TestPostgresDBSync(unittest.TestCase):
    def setUp(self):

        self.temp_schema = TEMP_SCHEMA
        self.db = Postgres()

        # Create a schema.
        setup_sql = f"""
                     DROP SCHEMA IF EXISTS {self.temp_schema} CASCADE;
                     CREATE SCHEMA {self.temp_schema};
                     """
        self.db.query(setup_sql)

        # Load dummy data to parsons tables
        self.table1 = Table.from_csv(f"{_dir}/test_data/sample_table_1.csv")
        self.table2 = Table.from_csv(f"{_dir}/test_data/sample_table_2.csv")

        # Create source table
        self.db.copy(self.table1, f"{self.temp_schema}.source")

        # Create DB Sync object
        self.db_sync = DBSync(self.db, self.db)

    def tearDown(self):
        # Drop the view, the table and the schema

        teardown_sql = f"""
                       DROP SCHEMA IF EXISTS {self.temp_schema} CASCADE;
                       """
        self.db.query(teardown_sql)

    def test_table_sync_full_drop(self):
        # Test a db sync with drop.

        self.db_sync.table_sync_full(
            f"{self.temp_schema}.source", f"{self.temp_schema}.destination"
        )

        source = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        destination = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(source, destination)

    def test_table_sync_full_truncate(self):
        # Test a db sync with truncate.

        self.db_sync.table_sync_full(
            f"{self.temp_schema}.source",
            f"{self.temp_schema}.destination",
            if_exists="truncate",
        )
        source = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        destination = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(source, destination)

    def test_table_sync_full_empty_table(self):
        # Test a full sync of a table when the source table is empty.

        # Empty the source table
        self.db.query(f"TRUNCATE {self.temp_schema}.source")

        # Attempt to sync.
        self.db_sync.table_sync_full(
            f"{self.temp_schema}.source", f"{self.temp_schema}.destination"
        )

    def test_table_sync_full_chunk(self):
        # Test chunking in full sync.

        self.db_sync.chunk_size = 10
        self.db_sync.table_sync_full(
            f"{self.temp_schema}.source", f"{self.temp_schema}.destination"
        )

        source = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        destination = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(source, destination)

    def test_table_sync_incremental(self):
        # Test that incremental sync

        self.db.copy(self.table1, f"{self.temp_schema}.destination")
        self.db.copy(self.table2, f"{self.temp_schema}.source", if_exists="append")
        self.db_sync.table_sync_incremental(
            f"{self.temp_schema}.source", f"{self.temp_schema}.destination", "pk"
        )

        count1 = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        count2 = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(count1, count2)

    def test_table_sync_incremental_chunk(self):
        # Test chunking of incremental sync.

        self.db_sync.chunk_size = 10
        self.db.copy(self.table1, f"{self.temp_schema}.destination")
        self.db.copy(self.table2, f"{self.temp_schema}.source", if_exists="append")
        self.db_sync.table_sync_incremental(
            f"{self.temp_schema}.source", f"{self.temp_schema}.destination", "pk"
        )

        count1 = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        count2 = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(count1, count2)

    def test_table_sync_incremental_create_destination_table(self):
        # Test that an incremental sync works if the destination table does not exist.

        self.db_sync.table_sync_incremental(
            f"{self.temp_schema}.source", f"{self.temp_schema}.destination", "pk"
        )

        count1 = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        count2 = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(count1, count2)

    def test_table_sync_incremental_empty_table(self):
        # Test an incremental sync of a table when the source table is empty.

        # Empty the source table
        self.db.query(f"TRUNCATE {self.temp_schema}.source")

        # Attempt to sync.
        self.db_sync.table_sync_incremental(
            f"{self.temp_schema}.source", f"{self.temp_schema}.destination", "pk"
        )


# These tests interact directly with the Postgres database. In order to run, set the
# env to LIVE_TEST='TRUE'.
@unittest.skipIf(
    not os.environ.get("LIVE_TEST"), "Skipping because not running live test"
)
class TestRedshiftDBSync(TestPostgresDBSync):
    """This test inherits all of the tests from the Postgres test."""

    def setUp(self):

        self.temp_schema = TEMP_SCHEMA
        self.db = Redshift()

        # Create a schema.
        setup_sql = f"""
                     DROP SCHEMA IF EXISTS {self.temp_schema} CASCADE;
                     CREATE SCHEMA {self.temp_schema};
                     """
        self.db.query(setup_sql)

        # Load dummy data to parsons tables
        self.table1 = Table.from_csv(f"{_dir}/test_data/sample_table_1.csv")
        self.table2 = Table.from_csv(f"{_dir}/test_data/sample_table_2.csv")

        # Create source table
        self.db.copy(self.table1, f"{self.temp_schema}.source")

        # Create DB Sync object
        self.db_sync = DBSync(self.db, self.db)


class TestFakeDBSync(unittest.TestCase):
    def setUp(self):
        self.fake_source = FakeDatabase()
        self.fake_destination = FakeDatabase()

    def test_table_sync_full(self):
        dbsync = DBSync(self.fake_source, self.fake_destination)
        source_data = Table(
            [
                {"id": 1, "value": 11},
                {"id": 2, "value": 121142},
                {"id": 3, "value": 111},
                {"id": 4, "value": 12211},
                {"id": 5, "value": 1231},
            ]
        )
        self.fake_source.setup_table("source", source_data)

        dbsync.table_sync_full("source", "destination")

        destination = self.fake_destination.table("destination")

        # Make sure the data came through
        assert_matching_tables(source_data, destination.data)

    def test_table_sync_incremental(self):
        dbsync = DBSync(self.fake_source, self.fake_destination)
        source_data = Table(
            [
                {"id": 1, "value": 11},
                {"id": 2, "value": 121142},
                {"id": 3, "value": 111},
                {"id": 4, "value": 12211},
                {"id": 5, "value": 1231},
            ]
        )
        self.fake_source.setup_table("source", source_data)

        # Start with one row
        destination_data = Table(
            [
                {"id": 1, "value": 11},
            ]
        )
        self.fake_destination.setup_table("destination", destination_data)

        dbsync.table_sync_incremental("source", "destination", "id")

        destination = self.fake_destination.table("destination")

        # Make sure the rest of the data came through
        assert_matching_tables(source_data, destination.data)

    def test_table_sync_full_with_retry(self):
        # Setup the dbsync with two retries
        dbsync = DBSync(self.fake_source, self.fake_destination, retries=2)
        source_data = Table(
            [
                {"id": 1, "value": 11},
                {"id": 2, "value": 121142},
            ]
        )
        self.fake_source.setup_table("source", source_data)

        # Have the copy fail twice
        self.fake_destination.setup_table("destination", Table(), failures=2)

        dbsync.table_sync_full("source", "destination")

        destination = self.fake_destination.table("destination")

        # Make sure all of the data still came through
        assert_matching_tables(source_data, destination.data)

    def test_table_sync_full_without_retry(self):
        # Setup the dbsync with no retries
        dbsync = DBSync(self.fake_source, self.fake_destination, retries=0)
        source_data = Table(
            [
                {"id": 1, "value": 11},
                {"id": 2, "value": 121142},
            ]
        )
        self.fake_source.setup_table("source", source_data)

        # Have the copy fail once
        self.fake_destination.setup_table("destination", Table(), failures=1)

        # Make sure the sync results in an exception
        self.assertRaises(
            ValueError, lambda: dbsync.table_sync_full("source", "destination")
        )

    def test_table_sync_full_order_by(self):
        dbsync = DBSync(self.fake_source, self.fake_destination)
        source_data = Table(
            [
                {"id": 1, "value": 21},
                {"id": 2, "value": 121142},
                {"id": 3, "value": 1},
            ]
        )
        self.fake_source.setup_table("source", source_data)
        self.fake_destination.setup_table("destination", Table())

        dbsync.table_sync_full("source", "destination", order_by="value")

        destination = self.fake_destination.table("destination")

        # Check that the rows were inserted in the expected order
        self.assertEqual(destination.data[0]["id"], 3)
        self.assertEqual(destination.data[1]["id"], 1)
        self.assertEqual(destination.data[2]["id"], 2)

    def test_table_sync_full_read_chunk(self):
        dbsync = DBSync(self.fake_source, self.fake_destination, read_chunk_size=2)
        source_data = Table(
            [
                {"id": 1, "value": 11},
                {"id": 2, "value": 121142},
                {"id": 3, "value": 111},
                {"id": 4, "value": 12211},
                {"id": 5, "value": 1231},
            ]
        )
        self.fake_source.setup_table("source", source_data)

        dbsync.table_sync_full("source", "destination")

        destination = self.fake_destination.table("destination")

        # Make sure the data came through
        assert_matching_tables(source_data, destination.data)

        # Make sure copy was called the expected number of times
        # read chunks of 2, 5 rows to write.. should be 3 copy calls
        self.assertEqual(
            len(self.fake_destination.copy_call_args),
            3,
            self.fake_destination.copy_call_args,
        )

    def test_table_sync_full_write_chunk(self):
        dbsync = DBSync(
            self.fake_source,
            self.fake_destination,
            read_chunk_size=1,
            write_chunk_size=3,
        )
        source_data = Table(
            [
                {"id": 1, "value": 11},
                {"id": 2, "value": 121142},
                {"id": 3, "value": 111},
                {"id": 4, "value": 12211},
                {"id": 5, "value": 1231},
            ]
        )
        self.fake_source.setup_table("source", source_data)

        dbsync.table_sync_full("source", "destination")

        destination = self.fake_destination.table("destination")

        # Make sure the data came through
        assert_matching_tables(source_data, destination.data)

        # Make sure copy was called the expected number of times
        # write chunks of 3, 5 rows to write.. should be 2 copy calls
        self.assertEqual(
            len(self.fake_destination.copy_call_args),
            2,
            self.fake_destination.copy_call_args,
        )



list_json = [
    {
        "number": "43-0000",
        "name": "Precinct 01 Turf 01",
        "eventSignups": None,
        "listSize": 62,
    },
    {
        "number": "44-0000",
        "name": "Precinct 01 Turf 02",
        "eventSignups": None,
        "listSize": 54,
    },
    {
        "number": "45-0000",
        "name": "Precinct 01 Turf 03",
        "eventSignups": None,
        "listSize": 70,
    },
    {
        "number": "46-0000",
        "name": "Precinct 01 Turf 04",
        "eventSignups": None,
        "listSize": 66,
    },
    {
        "number": "48-0000",
        "name": "Precinct 02 Turf 01",
        "eventSignups": None,
        "listSize": 45,
    },
    {
        "number": "49-0000",
        "name": "Precinct 02 Turf 02",
        "eventSignups": None,
        "listSize": 33,
    },
    {
        "number": "50-0000",
        "name": "Precinct 02 Turf 03",
        "eventSignups": None,
        "listSize": 63,
    },
    {
        "number": "51-0000",
        "name": "Precinct 02 Turf 04",
        "eventSignups": None,
        "listSize": 66,
    },
    {
        "number": "52-0000",
        "name": "Precinct 02 Turf 05",
        "eventSignups": None,
        "listSize": 39,
    },
    {
        "number": "55-0000",
        "name": "Precinct 03 Turf 01",
        "eventSignups": None,
        "listSize": 53,
    },
    {
        "number": "56-0000",
        "name": "Precinct 03 Turf 02",
        "eventSignups": None,
        "listSize": 45,
    },
    {
        "number": "57-0000",
        "name": "Precinct 03 Turf 03",
        "eventSignups": None,
        "listSize": 50,
    },
    {
        "number": "58-0000",
        "name": "Precinct 03 Turf 04",
        "eventSignups": None,
        "listSize": 54,
    },
    {
        "number": "59-0000",
        "name": "Precinct 03 Turf 05",
        "eventSignups": None,
        "listSize": 65,
    },
]

single_list_json = {
    "number": "43-0000",
    "name": "Precinct 01 Turf 01",
    "eventSignups": [],
    "listSize": 62,
}



from test.utils import mark_live_test
from parsons import Table


#####

START_DATE = "2020-01-01"
END_DATE = "2022-12-31"
EXPAND = True
LOWER_LIMIT = 1

# TODO: Invoke this, it should fail as 2000 is the max limit for
# all of the relevant events functions
UPPER_LIMIT = 2001


@mark_live_test
def test_get_calendars(live_pdi):
    response = live_pdi.get_calendars()

    assert type(response) == Table


@mark_live_test
def test_get_calendars_with_limit(live_pdi):
    response = live_pdi.get_calendars(limit=LOWER_LIMIT)

    assert response.num_rows == 1


@mark_live_test
def test_get_event_activities(live_pdi):
    response = live_pdi.get_event_activities(start_date=START_DATE, end_date=END_DATE)

    assert type(response) == Table


@mark_live_test
def test_get_event_activities_with_limit(live_pdi):
    response = live_pdi.get_event_activities(
        start_date=START_DATE, end_date=END_DATE, limit=LOWER_LIMIT
    )

    assert response.num_rows == 1


@mark_live_test
def test_get_event_activity_assignments(live_pdi):
    response = live_pdi.get_event_activity_assignments(
        start_date=START_DATE, end_date=END_DATE, expand=EXPAND
    )

    assert type(response) == Table


@mark_live_test
def test_get_event_activity_assignments_with_limit(live_pdi):
    response = live_pdi.get_event_activity_assignments(
        start_date=START_DATE, end_date=END_DATE, expand=EXPAND
    )

    assert response.num_rows == 1



import pickle
from typing import Optional, Union
import uuid

from google.cloud import bigquery
from google.cloud.bigquery import dbapi
from google.cloud.bigquery.job import LoadJobConfig
from google.cloud import exceptions
import petl

from parsons.databases.table import BaseTable
from parsons.databases.database_connector import DatabaseConnector
from parsons.etl import Table
from parsons.google.utitities import setup_google_application_credentials
from parsons.google.google_cloud_storage import GoogleCloudStorage
from parsons.utilities import check_env
from parsons.utilities.files import create_temp_file

BIGQUERY_TYPE_MAP = {
    "str": "STRING",
    "float": "FLOAT",
    "int": "INTEGER",
    "bool": "BOOLEAN",
    "datetime.datetime": "DATETIME",
    "datetime.date": "DATE",
    "datetime.time": "TIME",
    "dict": "RECORD",
    "NoneType": "STRING",
}

# Max number of rows that we query at a time, so we can avoid loading huge
# data sets into memory.
# 100k rows per batch at ~1k bytes each = ~100MB per batch.
QUERY_BATCH_SIZE = 100000


def get_table_ref(client, table_name):
    # Helper function to build a TableReference for our table
    parsed = parse_table_name(table_name)
    dataset_ref = client.dataset(parsed["dataset"])
    return dataset_ref.table(parsed["table"])


def parse_table_name(table_name):
    # Helper function to parse out the different components of a table ID
    parts = table_name.split(".")
    parts.reverse()
    parsed = {
        "project": None,
        "dataset": None,
        "table": None,
    }
    if len(parts) > 0:
        parsed["table"] = parts[0]
    if len(parts) > 1:
        parsed["dataset"] = parts[1]
    if len(parts) > 2:
        parsed["project"] = parts[2]
    return parsed


class GoogleBigQuery(DatabaseConnector):
    """
    Class for querying BigQuery table and returning the data as Parsons tables.

    This class requires application credentials in the form of a json. It can be passed
    in the following ways:

    * Set an environmental variable named ``GOOGLE_APPLICATION_CREDENTIALS`` with the
      local path to the credentials json.

      Example: ``GOOGLE_APPLICATION_CREDENTALS='path/to/creds.json'``

    * Pass in the path to the credentials using the ``app_creds`` argument.

    * Pass in a json string using the ``app_creds`` argument.

    Args:
        app_creds: str
            A credentials json string or a path to a json file. Not required
            if ``GOOGLE_APPLICATION_CREDENTIALS`` env variable set.
        project: str
            The project which the client is acting on behalf of. If not passed
            then will use the default inferred environment.
        location: str
            Default geographic location for tables
    """

    def __init__(self, app_creds=None, project=None, location=None):
        self.app_creds = app_creds

        setup_google_application_credentials(app_creds)

        self.project = project
        self.location = location

        # We will not create the client until we need to use it, since creating the client
        # without valid GOOGLE_APPLICATION_CREDENTIALS raises an exception.
        # This attribute will be used to hold the client once we have created it.
        self._client = None

        self._dbapi = dbapi

        self.dialect = "bigquery"

    def copy(
        self,
        tbl: Table,
        table_name: str,
        if_exists: str = "fail",
        tmp_gcs_bucket: Optional[str] = None,
        gcs_client: Optional[GoogleCloudStorage] = None,
        job_config: Optional[LoadJobConfig] = None,
        **load_kwargs,
    ):
        """
        Copy a :ref:`parsons-table` into Google BigQuery via Google Cloud Storage.

        `Args:`
            table_obj: obj
                The Parsons Table to copy into BigQuery.
            table_name: str
                The table name to load the data into.
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            tmp_gcs_bucket: str
                The name of the Google Cloud Storage bucket to use to stage the data to load
                into BigQuery. Required if `GCS_TEMP_BUCKET` is not specified.
            gcs_client: object
                The GoogleCloudStorage Connector to use for loading data into Google Cloud Storage.
            job_config: object
                A LoadJobConfig object to provide to the underlying call to load_table_from_uri
                on the BigQuery client. The function will create its own if not provided.
            **load_kwargs: kwargs
                Arguments to pass to the underlying load_table_from_uri call on the BigQuery
                client.
        """
        tmp_gcs_bucket = check_env.check("GCS_TEMP_BUCKET", tmp_gcs_bucket)

        if if_exists not in ["fail", "truncate", "append", "drop"]:
            raise ValueError(
                f"Unexpected value for if_exists: {if_exists}, must be one of "
                '"append", "drop", "truncate", or "fail"'
            )

        table_exists = self.table_exists(table_name)

        if not job_config:
            job_config = bigquery.LoadJobConfig()

        if not job_config.schema:
            job_config.schema = self._generate_schema(tbl)

        if not job_config.create_disposition:
            job_config.create_disposition = bigquery.CreateDisposition.CREATE_IF_NEEDED
        job_config.skip_leading_rows = 1
        job_config.source_format = bigquery.SourceFormat.CSV
        job_config.write_disposition = bigquery.WriteDisposition.WRITE_EMPTY

        if table_exists:
            if if_exists == "fail":
                raise ValueError("Table already exists.")
            elif if_exists == "drop":
                self.delete_table(table_name)
            elif if_exists == "append":
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND
            elif if_exists == "truncate":
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE

        gcs_client = gcs_client or GoogleCloudStorage()
        temp_blob_name = f"{uuid.uuid4()}.csv"
        temp_blob_uri = gcs_client.upload_table(tbl, tmp_gcs_bucket, temp_blob_name)

        # load CSV from Cloud Storage into BigQuery
        table_ref = get_table_ref(self.client, table_name)
        try:
            load_job = self.client.load_table_from_uri(
                temp_blob_uri,
                table_ref,
                job_config=job_config,
                **load_kwargs,
            )
            load_job.result()
        finally:
            gcs_client.delete_blob(tmp_gcs_bucket, temp_blob_name)

    def delete_table(self, table_name):
        """
        Delete a BigQuery table.

        `Args:`
            table_name: str
                The name of the table to delete.
        """
        table_ref = get_table_ref(self.client, table_name)
        self.client.delete_table(table_ref)

    def query(
        self, sql: str, parameters: Optional[Union[list, dict]] = None
    ) -> Optional[Table]:
        """
        Run a BigQuery query and return the results as a Parsons table.

        To include python variables in your query, it is recommended to pass them as parameters,
        following the BigQuery style where parameters are prefixed with `@`s.
        Using the ``parameters`` argument ensures that values are escaped properly, and avoids SQL
        injection attacks.

        **Parameter Examples**

        .. code-block:: python

           name = "Beatrice O'Brady"
           sql = 'SELECT * FROM my_table WHERE name = %s'
           rs.query(sql, parameters=[name])

        .. code-block:: python

           name = "Beatrice O'Brady"
           sql = "SELECT * FROM my_table WHERE name = %(name)s"
           rs.query(sql, parameters={'name': name})

        `Args:`
            sql: str
                A valid BigTable statement
            parameters: dict
                A dictionary of query parameters for BigQuery.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        # get our connection and cursor
        cursor = self._dbapi.connect(self.client).cursor()

        # Run the query
        cursor.execute(sql, parameters)

        # We will use a temp file to cache the results so that they are not all living
        # in memory. We'll use pickle to serialize the results to file in order to maintain
        # the proper data types (e.g. integer).
        temp_filename = create_temp_file()

        wrote_header = False
        with open(temp_filename, "wb") as temp_file:
            # Track whether we got data, since if we don't get any results we need to return None
            got_results = False
            while True:
                batch = cursor.fetchmany(QUERY_BATCH_SIZE)
                if len(batch) == 0:
                    break

                got_results = True

                for row in batch:
                    # Make sure we write out the header once and only once
                    if not wrote_header:
                        wrote_header = True
                        header = list(row.keys())
                        pickle.dump(header, temp_file)

                    row_data = list(row.values())
                    pickle.dump(row_data, temp_file)

        if not got_results:
            return None

        ptable = petl.frompickle(temp_filename)
        final_table = Table(ptable)

        return final_table

    def table_exists(self, table_name: str) -> bool:
        """
        Check whether or not the Google BigQuery table exists in the specified dataset.

        `Args:`
            table_name: str
                The name of the BigQuery table to check for
        `Returns:`
            bool
                True if the table exists in the specified dataset, false otherwise
        """
        table_ref = get_table_ref(self.client, table_name)
        try:
            self.client.get_table(table_ref)
        except exceptions.NotFound:
            return False

        return True

    @property
    def client(self):
        """
        Get the Google BigQuery client to use for making queries.

        `Returns:`
            `google.cloud.bigquery.client.Client`
        """
        if not self._client:
            # Create a BigQuery client to use to make the query
            self._client = bigquery.Client(project=self.project, location=self.location)

        return self._client

    def _generate_schema(self, tbl):
        stats = tbl.get_columns_type_stats()
        fields = []
        for stat in stats:
            petl_types = stat["type"]
            best_type = "str" if "str" in petl_types else petl_types[0]
            field_type = self._bigquery_type(best_type)
            field = bigquery.schema.SchemaField(stat["name"], field_type)
            fields.append(field)
        return fields

    @staticmethod
    def _bigquery_type(tp):
        return BIGQUERY_TYPE_MAP[tp]

    def table(self, table_name):
        # Return a MySQL table object

        return BigQueryTable(self, table_name)


class BigQueryTable(BaseTable):
    # BigQuery table object.

    def drop(self, cascade=False):
        """
        Drop the table.
        """

        self.db.delete_table(self.table)

    def truncate(self):
        """
        Truncate the table.
        """
        # BigQuery does not support truncate natively, so we will "load" an empty dataset
        # with write disposition of "truncate"
        table_ref = get_table_ref(self.db.client, self.table)
        bq_table = self.db.client.get_table(table_ref)

        # BigQuery wants the schema when we load the data, so we will grab it from the table
        job_config = bigquery.LoadJobConfig()
        job_config.schema = bq_table.schema

        empty_table = Table([])
        self.db.copy(
            empty_table, self.table, if_exists="truncate", job_config=job_config
        )



from datetime import datetime, timedelta
import re

from parsons.etl.table import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector


class Shopify(object):
    """
    Instantiate the Shopify class
    `Args:`
        subdomain: str
            The Shopify subdomain (e.g. ``myorg`` for myorg.myshopify.com) Not required if
            ``SHOPIFY_SUBDOMAIN`` env variable set.
        password: str
            The Shopify account password. Not required if ``SHOPIFY_PASSWORD`` env
            variable set.
        api_key: str
            The Shopify account API key. Not required if ``SHOPIFY_API_KEY`` env variable
            set.
        api_version: str
            The Shopify API version. Not required if ``SHOPIFY_API_VERSION`` env variable
            set.
        access_token: str
            The Shopify access token.  Not required if ``SHOPIFY_ACCESS_TOKEN`` env
            variable set. If argument or env variable is set, password and api_key
            are ignored.
    `Returns:`
        Shopify Class
    """

    def __init__(
        self,
        subdomain=None,
        password=None,
        api_key=None,
        api_version=None,
        access_token=None,
    ):
        self.subdomain = check_env.check("SHOPIFY_SUBDOMAIN", subdomain)
        self.access_token = check_env.check(
            "SHOPIFY_ACCESS_TOKEN", access_token, optional=True
        )
        self.password = check_env.check("SHOPIFY_PASSWORD", password, optional=True)
        self.api_key = check_env.check("SHOPIFY_API_KEY", api_key, optional=True)
        self.api_version = check_env.check("SHOPIFY_API_VERSION", api_version)
        self.base_url = "https://%s.myshopify.com/admin/api/%s/" % (
            self.subdomain,
            self.api_version,
        )
        if self.access_token is None and (
            self.password is None or self.api_key is None
        ):
            raise KeyError("Must set either access_token or both api_key and password.")
        if self.access_token is not None:
            self.client = APIConnector(
                self.base_url, headers={"X-Shopify-Access-Token": access_token}
            )
        else:
            self.client = APIConnector(
                self.base_url, auth=(self.api_key, self.password)
            )

    def get_count(self, query_date=None, since_id=None, table_name=None):
        """
        Get the count of rows in a table.
        `Args:`
            query_date: str
                Filter query by a date that rows were created. This filter is ignored if value
                is None.
            since_id: str
                Filter query by a minimum ID. This filter is ignored if value is None.
            table_name: str
                The name of the Shopify table to query.
        `Returns:`
            int
        """
        return (
            self.client.request(
                self.get_query_url(query_date, since_id, table_name), "GET"
            )
            .json()
            .get("count", 0)
        )

    def get_orders(self, query_date=None, since_id=None, completed=True):
        """
        Get Shopify orders.
        `Args:`
            query_date: str
                Filter query by a date that rows were created. Format: yyyy-mm-dd. This filter
                is ignored if value is None.
            since_id: str
                Filter query by a minimum ID. This filter is ignored if value is None.
            completed: bool
                True if only getting completed orders, False otherwise.
        `Returns:`
            Table Class
        """
        orders = []

        def _append_orders(url):
            nonlocal orders

            if completed:
                url += "&financial_status=paid"

            res = self.client.request(url, "GET")

            cur_orders = res.json().get("orders", [])

            # Flatten orders to non-complex types
            for order in cur_orders:
                keys_to_add = {}
                keys_to_delete = []

                for key1 in order:
                    if isinstance(order[key1], dict):
                        for key2 in order[key1]:
                            keys_to_add[key1 + "_" + key2] = order[key1][key2]
                        keys_to_delete.append(key1)
                    elif key1 == "note_attributes":
                        for note in order[key1]:
                            keys_to_add[key1 + "_" + note["name"]] = note["value"]

                order.update(keys_to_add)
                for key in keys_to_delete:
                    del order[key]

            orders += cur_orders

            return res

        res = _append_orders(self.get_query_url(query_date, since_id, "orders", False))

        # Get next page
        while res.headers.get("Link"):
            link = re.split("; |, ", res.headers.get("Link"))
            if len(link) and link[len(link) - 1] == 'rel="next"':
                res = _append_orders(link[len(link) - 2][1:-1])
            else:
                break

        return Table(orders)

    def get_query_url(
        self, query_date=None, since_id=None, table_name=None, count=True
    ):
        """
        Get the URL of a Shopify API request
        `Args:`
            query_date: str
                Filter query by a date that rows were created. Format: yyyy-mm-dd. This filter
                is ignored if value is None.
            since_id: str
                Filter query by a minimum ID. This filter is ignored if value is None.
            table_name: str
                The name of the Shopify table to query.
            count: bool
                True if refund should be included in Table, False otherwise.
        `Returns:`
            str
        """
        filters = "limit=250&status=any"

        if count:
            table = table_name + "/count.json"
        else:
            table = table_name + ".json"

        if query_date:
            # Specific date if provided
            query_date = datetime.strptime(query_date, "%Y-%m-%d")
            max_date = query_date + timedelta(days=1)
            filters += "&created_at_min={}&created_at_max={}".format(
                query_date.isoformat(), max_date.isoformat()
            )
        elif since_id:
            # Since ID if provided
            filters += "&since_id=%s" % since_id

        return self.base_url + "%s?%s" % (table, filters)

    def graphql(self, query):
        """
        Make GraphQL request. Reference: https://shopify.dev/api/admin-graphql
        `Args:`
            query: str
                GraphQL query.
        `Returns:`
            dict
        """
        return (
            self.client.request(
                self.base_url + "graphql.json", "POST", json={"query": query}
            )
            .json()
            .get("data")
        )

    @classmethod
    def load_to_table(
        cls,
        subdomain=None,
        password=None,
        api_key=None,
        api_version=None,
        query_date=None,
        since_id=None,
        completed=True,
    ):
        """
        Fast classmethod so you can get the data all at once:
            tabledata = Shopify.load_to_table(subdomain='myorg', password='abc123',
                                            api_key='abc123', api_version='2020-10',
                                            query_date='2020-10-20', since_id='8414',
                                            completed=True)
        This instantiates the class and makes the appropriate query type to Shopify's orders
        table based on which arguments are supplied.
        `Args:`
            subdomain: str
                The Shopify subdomain (e.g. ``myorg`` for myorg.myshopify.com).
            password: str
                The Shopify account password.
            api_key: str
                The Shopify account API key.
            api_version: str
                The Shopify API version.
            query_date: str
                Filter query by a date that rows were created. Format: yyyy-mm-dd. This filter
                is ignored if value is None.
            since_id: str
                Filter query by a minimum ID. This filter is ignored if value is None.
            completed: bool
                True if only getting completed orders, False otherwise.
                value as value
        `Returns:`
            Table Class
        """
        return cls(subdomain, password, api_key, api_version).get_orders(
            query_date, since_id, completed
        )



from sqlalchemy import create_engine, Table, MetaData
import logging

logger = logging.getLogger(__name__)


class Alchemy:
    def generate_engine(self):
        """
        Generate a SQL Alchemy engine.
        """

        alchemy_url = self.generate_alchemy_url()
        return create_engine(alchemy_url, echo=False, convert_unicode=True)

    def generate_alchemy_url(self):
        """
        Generate a SQL Alchemy engine
        https://docs.sqlalchemy.org/en/14/core/engines.html#
        """

        if self.dialect == "redshift" or self.dialect == "postgres":
            connection_schema = "postgresql+psycopg2"
        elif self.dialect == "mysql":
            connection_schema = "mysql+mysqlconnector"

        params = [
            (self.username, self.username),
            (self.password, f":{self.password}"),
            (self.host, f"@{self.host}"),
            (self.port, f":{self.port}"),
            (self.db, f"/{self.db}"),
        ]

        url = f"{connection_schema}://"

        for i in params:
            if i[0]:
                url += i[1]

        return url

    def get_table_object(self, table_name):
        """
        Get a SQL Alchemy table object.
        """

        schema, table_name = self.split_table_name(table_name)
        db_meta = MetaData(bind=self.generate_engine(), schema=schema)
        return Table(table_name, db_meta, autoload=True)

    def create_table(self, table_object, table_name):
        """
        Create a table based on table object data.
        """

        schema, table_name = self.split_table_name(table_name)

        if schema:
            table_object.schema = schema
        if table_name:
            table_object.table_name = table_name

        table_object.metadata.create_all(self.generate_engine())

    @staticmethod
    def split_table_name(full_table_name):
        """
        Utility method to parse the schema and table name.
        """

        if "." not in full_table_name:
            return "public", full_table_name

        try:
            schema, table = full_table_name.split(".")
        except ValueError as e:
            if "too many values to unpack" in str(e):
                raise ValueError(f"Invalid database table {full_table_name}")

        return schema, table



test_agent = [
    {
        "available": False,
        "occasional": True,
        "id": 47020956237,
        "signature": '<div dir="ltr"><p><br></p>\n</div>',
        "ticket_scope": 1,
        "created_at": "2020-01-17T15:07:01Z",
        "updated_at": "2020-02-05T00:58:37Z",
        "available_since": None,
        "type": "support_agent",
        "contact": {
            "active": True,
            "email": "person@email.org",
            "job_title": None,
            "language": "en",
            "last_login_at": "2020-01-24T22:49:52Z",
            "mobile": None,
            "name": "Alissa",
            "phone": None,
            "time_zone": "Bogota",
            "created_at": "2020-01-17T15:07:00Z",
            "updated_at": "2020-01-24T22:49:52Z",
        },
    }
]

test_ticket = [
    {
        "cc_emails": ["person@email.org", "person2@email.org"],
        "fwd_emails": [],
        "reply_cc_emails": ["person@email.org", "person2@email.org"],
        "ticket_cc_emails": ["person@email.org", "person2@email.org"],
        "fr_escalated": False,
        "spam": False,
        "email_config_id": None,
        "group_id": 47000643034,
        "priority": 1,
        "requester_id": 47021937449,
        "responder_id": 47017224681,
        "source": 3,
        "company_id": 47000491688,
        "status": 5,
        "subject": "My thing is broken.",
        "association_type": None,
        "to_emails": None,
        "product_id": None,
        "id": 84,
        "type": "Support Request 1",
        "due_by": "2020-02-19T22:00:00Z",
        "fr_due_by": "2020-02-06T17:00:00Z",
        "is_escalated": False,
        "custom_fields": {},
        "created_at": "2020-02-05T22:17:41Z",
        "updated_at": "2020-02-06T02:07:37Z",
        "associated_tickets_count": None,
        "tags": [],
    }
]

test_company = [
    {
        "id": 47000491701,
        "name": "Big Org",
        "description": None,
        "note": None,
        "domains": [],
        "created_at": "2020-01-09T20:43:09Z",
        "updated_at": "2020-01-09T20:43:09Z",
        "custom_fields": {},
        "health_score": None,
        "account_tier": "Tier 2",
        "renewal_date": "2020-12-31T00:00:00Z",
        "industry": None,
    }
]

test_contact = [
    {
        "active": False,
        "address": None,
        "company_id": 47000491686,
        "description": None,
        "email": "person@email.org",
        "id": 47021299020,
        "job_title": None,
        "language": "en",
        "mobile": None,
        "name": "Percy Person",
        "phone": "N/A",
        "time_zone": "Bogota",
        "twitter_id": None,
        "custom_fields": {},
        "facebook_id": None,
        "created_at": "2020-01-27T16:44:34Z",
        "updated_at": "2020-01-27T16:44:34Z",
        "unique_external_id": None,
    }
]



import unittest
import unittest.mock
from test.utils import assert_matching_tables

import requests_mock
from parsons import Auth0, Table

CLIENT_ID = "abc"
CLIENT_SECRET = "def"
DOMAIN = "fakedomain.auth0.com"


class TestAuth0(unittest.TestCase):
    def setUp(self):
        self.auth0 = Auth0(CLIENT_ID, CLIENT_SECRET, DOMAIN)
        self.fake_upsert_person = {
            "email": "fakeemail@fakedomain.com",
            "given_name": "Fakey",
            "family_name": "McFakerson",
            "username": "fakeusername",
            "user_id": 3,
        }

    @requests_mock.Mocker()
    def test_delete_user(self, m):
        user_id = 1
        m.delete(f"{self.auth0.base_url}/api/v2/users/{user_id}", status_code=204)
        self.assertEqual(self.auth0.delete_user(user_id), 204)

    @requests_mock.Mocker()
    def test_get_users_by_email(self, m):
        email = "fakeemail@fakedomain.com"
        mock_users = [{"email": "fake3mail@fakedomain.com", "id": 2}]
        m.get(
            f"{self.auth0.base_url}/api/v2/users-by-email?email={email}",
            json=mock_users,
        )
        assert_matching_tables(
            self.auth0.get_users_by_email(email), Table(mock_users), True
        )

    @requests_mock.Mocker()
    def test_upsert_user(self, m):
        user = self.fake_upsert_person
        email = user["email"]
        m.get(
            f"{self.auth0.base_url}/api/v2/users-by-email?email={email}",
            json=[user],
        )
        mock_resp = unittest.mock.MagicMock()
        mock_resp.status_code = 200
        m.patch(f"{self.auth0.base_url}/api/v2/users/{user['user_id']}", [mock_resp])
        m.post(f"{self.auth0.base_url}/api/v2/users", mock_resp)
        ret = self.auth0.upsert_user(
            email,
            user["username"],
            user["given_name"],
            user["family_name"],
            {},
            {},
        )
        self.assertEqual(ret.status_code, 200)



import os
from parsons.aws.s3 import S3
import time
import logging

logger = logging.getLogger(__name__)

S3_TEMP_KEY_PREFIX = "Parsons_RedshiftCopyTable"


class RedshiftCopyTable(object):

    aws_access_key_id = None
    aws_secret_access_key = None
    iam_role = None

    def __init__(self, use_env_token=True):
        self.use_env_token = use_env_token

    def copy_statement(
        self,
        table_name,
        bucket,
        key,
        manifest=False,
        data_type="csv",
        csv_delimiter=",",
        max_errors=0,
        statupdate=None,
        compupdate=None,
        ignoreheader=1,
        acceptanydate=True,
        dateformat="auto",
        timeformat="auto",
        emptyasnull=True,
        blanksasnull=True,
        nullas=None,
        acceptinvchars=True,
        truncatecolumns=False,
        specifycols=None,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        compression=None,
        bucket_region=None,
    ):

        # Source / Destination
        source = f"s3://{bucket}/{key}"

        # Add column list for mapping or if there are fewer columns on source file
        col_list = f"({', '.join(specifycols)})" if specifycols is not None else ""

        sql = f"copy {table_name}{col_list} \nfrom '{source}' \n"

        # Generate credentials
        sql += self.get_creds(aws_access_key_id, aws_secret_access_key)

        # Other options
        if manifest:
            sql += "manifest \n"
        if bucket_region:
            sql += f"region '{bucket_region}'\n"
            logger.info(
                "Copying data from S3 bucket %s in region %s", bucket, bucket_region
            )
        sql += f"maxerror {max_errors} \n"

        # Redshift has some default behavior when statupdate is left out
        # vs when it is explicitly set as on or off.
        if statupdate is not None:
            if statupdate:
                sql += "statupdate on\n"
            else:
                sql += "statupdate off\n"

        # Redshift has some default behavior when compupdate is left out
        # vs when it is explicitly set as on or off.
        if compupdate is not None:
            if compupdate:
                sql += "compupdate on\n"
            else:
                sql += "compupdate off\n"

        if ignoreheader:
            sql += f"ignoreheader {ignoreheader} \n"
        if acceptanydate:
            sql += "acceptanydate \n"
        sql += f"dateformat '{dateformat}' \n"
        sql += f"timeformat '{timeformat}' \n"
        if emptyasnull:
            sql += "emptyasnull \n"
        if blanksasnull:
            sql += "blanksasnull \n"
        if nullas:
            sql += f"null as {nullas}"
        if acceptinvchars:
            sql += "acceptinvchars \n"
        if truncatecolumns:
            sql += "truncatecolumns \n"

        # Data Type
        if data_type == "csv":
            sql += f"csv delimiter '{csv_delimiter}' \n"
        else:
            raise TypeError("Invalid data type specified.")

        if compression == "gzip":
            sql += "gzip \n"

        sql += ";"

        return sql

    def get_creds(self, aws_access_key_id, aws_secret_access_key):

        if aws_access_key_id and aws_secret_access_key:
            # When we have credentials, then we don't need to set them again
            pass

        elif self.iam_role:
            # bail early, since the bottom is specifically formatted with creds
            return f"credentials 'aws_iam_role={self.iam_role}'\n"

        elif self.aws_access_key_id and self.aws_secret_access_key:

            aws_access_key_id = self.aws_access_key_id
            aws_secret_access_key = self.aws_secret_access_key

        elif (
            "AWS_ACCESS_KEY_ID" in os.environ and "AWS_SECRET_ACCESS_KEY" in os.environ
        ):

            aws_access_key_id = os.environ["AWS_ACCESS_KEY_ID"]
            aws_secret_access_key = os.environ["AWS_SECRET_ACCESS_KEY"]

        else:

            s3 = S3(use_env_token=self.use_env_token)
            creds = s3.aws.session.get_credentials()
            aws_access_key_id = creds.access_key
            aws_secret_access_key = creds.secret_key

        return "credentials 'aws_access_key_id={};aws_secret_access_key={}'\n".format(
            aws_access_key_id, aws_secret_access_key
        )

    def temp_s3_copy(
        self,
        tbl,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        csv_encoding="utf-8",
    ):

        if not self.s3_temp_bucket:
            raise KeyError(
                (
                    "Missing S3_TEMP_BUCKET, needed for transferring data to Redshift. "
                    "Must be specified as env vars or kwargs"
                )
            )

        # Coalesce S3 Key arguments
        aws_access_key_id = aws_access_key_id or self.aws_access_key_id
        aws_secret_access_key = aws_secret_access_key or self.aws_secret_access_key

        self.s3 = S3(
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            use_env_token=self.use_env_token,
        )

        hashed_name = hash(time.time())
        key = f"{S3_TEMP_KEY_PREFIX}/{hashed_name}.csv.gz"
        if self.s3_temp_bucket_prefix:
            key = self.s3_temp_bucket_prefix + "/" + key

        # Convert table to compressed CSV file, to optimize the transfers to S3 and to
        # Redshift.
        local_path = tbl.to_csv(temp_file_compression="gzip", encoding=csv_encoding)
        # Copy table to bucket
        self.s3.put_file(self.s3_temp_bucket, key, local_path)

        return key

    def temp_s3_delete(self, key):

        if key:
            self.s3.remove_file(self.s3_temp_bucket, key)



COL_NAME_MAX_LEN = 120

FLOAT = "float"

REPLACE_CHARS = {" ": ""}


# Max length of a Redshift VARCHAR column
VARCHAR_MAX = 65535

# List of varchar lengths to use for columns -- this list needs to be in order from smallest to
# largest
VARCHAR_STEPS = [32, 64, 128, 256, 512, 1024, 4096, 8192, 16384]



import os
import time

from parsons.etl.table import Table
from parsons.utilities.check_env import check

from slackclient import SlackClient
from slackclient.exceptions import SlackClientError

import requests


class Slack(object):
    def __init__(self, api_key=None):

        if api_key is None:

            try:
                self.api_key = os.environ["SLACK_API_TOKEN"]

            except KeyError:
                raise KeyError(
                    "Missing api_key. It must be passed as an "
                    "argument or stored as environmental variable"
                )

        else:

            self.api_key = api_key

        self.client = SlackClient(self.api_key)

    def channels(
        self, fields=["id", "name"], exclude_archived=False, types=["public_channel"]
    ):
        """
        Return a list of all channels in a Slack team.

        `Args:`
            fields: list
                A list of the fields to return. By default, only the channel
                `id` and `name` are returned. See
                https://api.slack.com/methods/conversations.list for a full
                list of available fields. `Notes:` nested fields are unpacked.
            exclude_archived: bool
                Set to `True` to exclude archived channels from the list.
                Default is false.
            types: list
                Mix and match channel types by providing a list of any
                combination of `public_channel`, `private_channel`,
                `mpim` (aka group messages), or `im` (aka 1-1 messages).
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        tbl = self._paginate_request(
            "conversations.list",
            "channels",
            types=types,
            exclude_archived=exclude_archived,
        )

        tbl.unpack_dict(
            "topic", include_original=False, prepend=True, prepend_value="topic"
        )
        tbl.unpack_dict(
            "purpose", include_original=False, prepend=True, prepend_value="purpose"
        )

        rm_cols = [x for x in tbl.columns if x not in fields]
        tbl.remove_column(*rm_cols)

        return tbl

    def users(
        self,
        fields=[
            "id",
            "name",
            "deleted",
            "profile_real_name_normalized",
            "profile_email",
        ],
    ):
        """
        Return a list of all users in a Slack team.

        `Args:`
            fields: list
                A list of the fields to return. By default, only the user
                `id` and `name` and `deleted` status are returned. See
                https://api.slack.com/methods/users.list for a full list of
                available fields. `Notes:` nested fields are unpacked.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = self._paginate_request("users.list", "members", include_locale=True)

        tbl.unpack_dict(
            "profile", include_original=False, prepend=True, prepend_value="profile"
        )

        rm_cols = [x for x in tbl.columns if x not in fields]
        tbl.remove_column(*rm_cols)

        return tbl

    @classmethod
    def message(cls, channel, text, webhook=None, parent_message_id=None):
        """
        Send a message to a Slack channel with a webhook instead of an api_key.
        You might not have the full-access API key but still want to notify a channel
        `Args:`
            channel: str
                The name or id of a `public_channel`, a `private_channel`, or
                an `im` (aka 1-1 message).
            text: str
                Text of the message to send.
            webhook: str
                If you have a webhook url instead of an api_key
                Looks like: https://hooks.slack.com/services/Txxxxxxx/Bxxxxxx/Dxxxxxxx
            parent_message_id: str
                The `ts` value of the parent message. If used, this will thread the message.
        """
        webhook = check("SLACK_API_WEBHOOK", webhook, optional=True)
        payload = {"channel": channel, "text": text}
        if parent_message_id:
            payload["thread_ts"] = parent_message_id
        return requests.post(webhook, json=payload)

    def message_channel(self, channel, text, as_user=False, parent_message_id=None):
        """
        Send a message to a Slack channel

        `Args:`
            channel: str
                The name or id of a `public_channel`, a `private_channel`, or
                an `im` (aka 1-1 message).
            text: str
                Text of the message to send.
            as_user: str
                Pass true to post the message as the authenticated user,
                instead of as a bot. Defaults to false. See
                https://api.slack.com/methods/chat.postMessage#authorship for
                more information about Slack authorship.
            parent_message_id: str
                The `ts` value of the parent message. If used, this will thread the message.
        `Returns:`
            `dict`:
                A response json
        """
        resp = self.client.api_call(
            "chat.postMessage",
            channel=channel,
            text=text,
            as_user=as_user,
            thread_ts=parent_message_id,
        )

        if not resp["ok"]:

            if resp["error"] == "ratelimited":
                time.sleep(int(resp["headers"]["Retry-After"]))

                resp = self.client.api_call(
                    "chat.postMessage", channel=channel, text=text, as_user=as_user
                )

            raise SlackClientError(resp["error"])

        return resp

    def upload_file(
        self,
        channels,
        filename,
        filetype=None,
        initial_comment=None,
        title=None,
        is_binary=False,
    ):
        """
        Upload a file to Slack channel(s).

        `Args:`
            channels: list
                The list of channel names or IDs where the file will be shared.
            filename: str
                The path to the file to be uploaded.
            filetype: str
                A file type identifier. If None, type will be inferred base on
                file extension. This is used to determine what fields are
                available for that object. See https://api.slack.com/types/file
                for a list of valid types and for more information about the
                file object.
            initial_comment: str
                The text of the message to send along with the file.
            title: str
                Title of the file to be uploaded.
            is_binary: bool
                If True, open this file in binary mode. This is needed if
                uploading binary files. Defaults to False.
        `Returns:`
            `dict`:
                A response json
        """
        if filetype is None and "." in filename:
            filetype = filename.split(".")[-1]

        mode = "rb" if is_binary else "r"
        with open(filename, mode) as file_content:
            resp = self.client.api_call(
                "files.upload",
                channels=channels,
                file=file_content,
                filetype=filetype,
                initial_comment=initial_comment,
                title=title,
            )

            if not resp["ok"]:

                if resp["error"] == "ratelimited":
                    time.sleep(int(resp["headers"]["Retry-After"]))

                    resp = self.client.api_call(
                        "files.upload",
                        channels=channels,
                        file=file_content,
                        filetype=filetype,
                        initial_comment=initial_comment,
                        title=title,
                    )

                raise SlackClientError(resp["error"])

        return resp

    def _paginate_request(self, endpoint, collection, **kwargs):
        # The max object we're requesting at a time.
        # This is an nternal limit to not overload slack api
        LIMIT = 200

        items = []
        next_page = True
        cursor = None
        while next_page:
            resp = self.client.api_call(endpoint, cursor=cursor, limit=LIMIT, **kwargs)

            if not resp["ok"]:

                if resp["error"] == "ratelimited":
                    time.sleep(int(resp["headers"]["Retry-After"]))
                    continue

                raise SlackClientError(resp["error"])

            items.extend(resp[collection])

            if resp["response_metadata"]["next_cursor"]:
                cursor = resp["response_metadata"]["next_cursor"]
            else:
                next_page = False

        return Table(items)



from parsons.databases.database.database import DatabaseCreateStatement
import parsons.databases.redshift.constants as consts

import petl
import logging

logger = logging.getLogger(__name__)


class RedshiftCreateTable(DatabaseCreateStatement):
    def __init__(self):
        super().__init__()

        self.COL_NAME_MAX_LEN = consts.COL_NAME_MAX_LEN
        self.REPLACE_CHARS = consts.REPLACE_CHARS

        # Currently smallints are coded as ints
        self.SMALLINT = self.INT
        # Redshift doesn't have a medium int
        self.MEDIUMINT = self.INT

        # Currently py floats are coded as Redshift decimals
        self.FLOAT = consts.FLOAT

        self.VARCHAR_MAX = consts.VARCHAR_MAX
        self.VARCHAR_STEPS = consts.VARCHAR_STEPS

    # the default behavior is f"{col}_"
    def _rename_reserved_word(self, col, index):
        """Return the renamed column.

        `Args`:
            col: str
                The column to rename.
            index: int
                (Optional) The index of the column.
        `Returns`:
            str
                The rename column.
        """
        return f"col_{index}"

    def create_statement(
        self,
        tbl,
        table_name,
        padding=None,
        distkey=None,
        sortkey=None,
        varchar_max=None,
        varchar_truncate=True,
        columntypes=None,
        strict_length=True,
    ):

        # Warn the user if they don't provide a DIST key or a SORT key
        self._log_key_warning(distkey=distkey, sortkey=sortkey, method="copy")

        # Generate a table create statement

        # Validate and rename column names if needed
        tbl.table = petl.setheader(tbl.table, self.column_name_validate(tbl.columns))

        if tbl.num_rows == 0:
            raise ValueError("Table is empty. Must have 1 or more rows.")

        mapping = self.generate_data_types(tbl)

        if padding:
            mapping["longest"] = self.vc_padding(mapping, padding)
        elif not strict_length:
            mapping["longest"] = self.vc_step(mapping)

        if varchar_max:
            mapping["longest"] = self.vc_max(mapping, varchar_max)

        if varchar_truncate:
            mapping["longest"] = self.vc_trunc(mapping)

        mapping["longest"] = self.vc_validate(mapping)

        # Add any provided column type overrides
        if columntypes:
            for i in range(len(mapping["headers"])):
                col = mapping["headers"][i]
                if columntypes.get(col):
                    mapping["type_list"][i] = columntypes[col]

        # Enclose in quotes
        mapping["headers"] = ['"{}"'.format(h) for h in mapping["headers"]]

        return self.create_sql(table_name, mapping, distkey=distkey, sortkey=sortkey)

    # This is for backwards compatability
    def data_type(self, val, current_type):
        return self.detect_data_type(val, current_type)

    # This is for backwards compatability
    def is_valid_integer(self, val):
        return self.is_valid_sql_num(val)

    def generate_data_types(self, table):
        # Generate column data types

        longest, type_list = [], []

        cont = petl.records(table.table)

        # Populate empty values for the columns
        for col in table.columns:
            longest.append(0)
            type_list.append("")

        for row in cont:
            for i in range(len(row)):
                # NA is the csv null value
                if type_list[i] == "varchar" or row[i] in ["NA", ""]:
                    pass
                else:
                    var_type = self.data_type(row[i], type_list[i])
                    type_list[i] = var_type

                # Calculate width
                if len(str(row[i]).encode("utf-8")) > longest[i]:
                    longest[i] = len(str(row[i]).encode("utf-8"))

        # In L138 'NA' and '' will be skipped
        # If the entire column is either one of those (or a mix of the two)
        # the type will be empty.
        # Fill with a default varchar
        type_list = [typ or "varchar" for typ in type_list]

        return {"longest": longest, "headers": table.columns, "type_list": type_list}

    def vc_padding(self, mapping, padding):
        # Pad the width of a varchar column

        return [int(c + (c * padding)) for c in mapping["longest"]]

    def vc_step(self, mapping):
        return [self.round_longest(c) for c in mapping["longest"]]

    def vc_max(self, mapping, columns):
        # Set the varchar width of a column to the maximum

        for c in columns:

            try:
                idx = mapping["headers"].index(c)
                mapping["longest"][idx] = self.VARCHAR_MAX

            except KeyError as error:
                logger.error("Could not find column name provided.")
                raise error

        return mapping["longest"]

    def vc_trunc(self, mapping):

        return [
            self.VARCHAR_MAX if c > self.VARCHAR_MAX else c for c in mapping["longest"]
        ]

    def vc_validate(self, mapping):

        return [1 if c == 0 else c for c in mapping["longest"]]

    def create_sql(self, table_name, mapping, distkey=None, sortkey=None):
        # Generate the sql to create the table

        statement = "create table {} (".format(table_name)

        for i in range(len(mapping["headers"])):
            if mapping["type_list"][i] == "varchar":
                statement = (statement + "\n  {} varchar({}),").format(
                    str(mapping["headers"][i]).lower(), str(mapping["longest"][i])
                )
            else:
                statement = (statement + "\n  " + "{} {}" + ",").format(
                    str(mapping["headers"][i]).lower(), mapping["type_list"][i]
                )

        statement = statement[:-1] + ") "

        if distkey:
            statement += "\ndistkey({}) ".format(distkey)

        if sortkey and isinstance(sortkey, list):
            statement += "\ncompound sortkey("
            statement += ", ".join(sortkey)
            statement += ")"
        elif sortkey:
            statement += "\nsortkey({})".format(sortkey)

        statement += ";"

        return statement

    # This is for backwards compatability
    def column_name_validate(self, columns):
        return self.format_columns(columns, col_prefix="col_")

    @staticmethod
    def _log_key_warning(distkey=None, sortkey=None, method=""):
        # Log a warning message advising the user about DIST and SORT keys

        if distkey and sortkey:
            return

        keys = [
            (
                distkey,
                "DIST",
                "https://aws.amazon.com/about-aws/whats-new/2019/08/amazon-redshift-"
                "now-recommends-distribution-keys-for-improved-query-performance/",
            ),
            (
                sortkey,
                "SORT",
                "https://docs.amazonaws.cn/en_us/redshift/latest/dg/c_best-practices-"
                "sort-key.html",
            ),
        ]
        warning = "".join(
            [
                "You didn't provide a {} key to method `parsons.redshift.Redshift.{}`.\n"
                "You can learn about best practices here:\n{}.\n".format(
                    keyname, method, keyinfo
                )
                for key, keyname, keyinfo in keys
                if not key
            ]
        )

        warning += "You may be able to further optimize your queries."

        logger.warning(warning)

    @staticmethod
    def round_longest(longest):
        # Find the value that will work best to fit our longest column value
        for step in consts.VARCHAR_STEPS:
            # Make sure we have padding
            if longest < step / 2:
                return step

        return consts.VARCHAR_MAX



import logging
import surveygizmo
from parsons.etl import Table
from parsons.utilities import check_env

logger = logging.getLogger(__name__)


def sg_compatibility():
    # Create backwards compatibility with SurveyGizmo class

    import os

    if os.getenv("SURVEYGIZMO_API_TOKEN"):
        os.environ["ALCHEMER_API_TOKEN"] = os.getenv("SURVEYGIZMO_API_TOKEN")

    if os.getenv("SURVEYGIZMO_API_TOKEN_SECRET"):
        os.environ["ALCHEMER_API_TOKEN_SECRET"] = os.getenv(
            "SURVEYGIZMO_API_TOKEN_SECRET"
        )

    if os.getenv("SURVEYGIZMO_API_VERSION"):
        os.environ["ALCHEMER_API_VERSION"] = os.getenv("SURVEYGIZMO_API_VERSION")


class Alchemer(object):
    """
    Instantiate Alchemer Class

    `Args:`
        api_token:
            The Alchemer-provided application token. Not required if
            ``ALCHEMER_API_TOKEN`` env variable set.

        api_token:
            The Alchemer-provided application token. Not required if
            ``ALCHEMER_API_TOKEN_SECRET`` env variable set.

        api_version:
            The version of the API that you would like to use. Not required if
            ``ALCHEMER_API_VERSION`` env variable set.
            Default v5

    `Returns:`
        Alchemer Class
    """

    def __init__(self, api_token=None, api_token_secret=None, api_version="v5"):

        sg_compatibility()

        self.api_token = check_env.check("ALCHEMER_API_TOKEN", api_token)
        self.api_token_secret = check_env.check(
            "ALCHEMER_API_TOKEN_SECRET", api_token_secret
        )
        self.api_version = check_env.check("ALCHEMER_API_VERSION", api_version)

        self._client = surveygizmo.SurveyGizmo(
            api_version=self.api_version,
            api_token=self.api_token,
            api_token_secret=self.api_token_secret,
        )

    def get_surveys(self, page=None):
        """
        Get a table of lists under the account.

        `Args:`
            page : int
                Retrieve a specific page of responses. If not given,
                then all pages are retrieved.

        `Returns:`
            Table Class
        """

        r = self._client.api.survey.list(page)
        data = r["data"]

        if not page:
            while r["page"] < r["total_pages"]:
                r = self._client.api.survey.list(page=(r["page"] + 1))
                data.extend(r["data"])

        tbl = Table(data).remove_column("links")
        tbl.unpack_dict("statistics", prepend=False)

        logger.info(f"Found {tbl.num_rows} surveys.")

        return tbl

    def get_survey_responses(self, survey_id, page=None):
        """
        Get the responses for a given survey.

        `Args:`
            survey_id: string
                The id of survey for which to retrieve the responses.

            page : int
                Retrieve a specific page of responses. If not given,
                then all pages are retrieved.

        `Returns:`
            Table Class
        """

        r = self._client.api.surveyresponse.list(survey_id, page)
        logger.info(f"{survey_id}: {r['total_count']} responses.")
        data = r["data"]

        if not page:
            while r["page"] < r["total_pages"]:
                r = self._client.api.surveyresponse.list(
                    survey_id, page=(r["page"] + 1)
                )
                data.extend(r["data"])

        tbl = Table(data).add_column("survey_id", survey_id, index=1)

        logger.info(f"Found #{tbl.num_rows} responses.")

        return tbl


# Backwards compatibility for old SurveyGizmo class.
SurveyGizmo = Alchemer



import unittest
import os
from parsons import CivisClient, Table

# from . import scratch_creds


@unittest.skipIf(
    not os.environ.get("LIVE_TEST"), "Skipping because not running live test"
)
class TestCivisClient(unittest.TestCase):
    def setUp(self):

        self.civis = CivisClient()

        # Create a schema, create a table, create a view
        setup_sql = """
                    drop schema if exists test_parsons cascade;
                    create schema test_parsons;
                    """

        self.lst_dicts = [{"first": "Bob", "last": "Smith"}]
        self.tbl = Table(self.lst_dicts)

        self.civis.query(setup_sql)

    def tearDown(self):

        # Drop the view, the table and the schema
        teardown_sql = """
                       drop schema if exists test_parsons cascade;
                       """

        self.civis.query(teardown_sql)

    def test_table_import_query(self):

        # Test that a good table imports correctly
        self.civis.table_import(self.tbl, "test_parsons.test_table")

    def test_query(self):

        # Test that queries match
        self.civis.table_import(self.tbl, "test_parsons.test_table")
        tbl = self.civis.query("SELECT COUNT(*) FROM test_parsons.test_table")
        self.assertEqual(tbl[0]["count"], "1")

    def test_to_civis(self):

        # Test that the to_civis() method works too
        self.tbl.to_civis("test_parsons.test_table")



from parsons.etl.table import Table
from parsons.utilities import check_env
from parsons.utilities.oauth_api_connector import OAuth2APIConnector


class Controlshift(object):
    """
    Instantiate the Controlshift class. Requires an API Application integration.
    For more info on setup, see:
    https://developers.controlshiftlabs.com/#authenticated-rest-api-quickstart-guide

    `Args:`
        hostname: str
            The URL for the homepage/login page of the organization's Controlshift
            instance (e.g. demo.controlshift.app). Not required if
            ``CONTROLSHIFT_HOSTNAME`` env variable is set.
        client_id: str
            The Client ID for your REST API Application. Not required if
            ``CONTROLSHIFT_CLIENT_ID`` env variable is set.
        client_secret: str
            The Client Secret for your REST API Application. Not required if
            ``CONTROLSHIFT_CLIENT_SECRET`` env variable is set.
    `Returns:`
        Controlshift Class
    """

    def __init__(self, hostname=None, client_id=None, client_secret=None):

        self.hostname = check_env.check("CONTROLSHIFT_HOSTNAME", hostname)
        token_url = f"{self.hostname}/oauth/token"
        self.client = OAuth2APIConnector(
            self.hostname,
            client_id=check_env.check("CONTROLSHIFT_CLIENT_ID", client_id),
            client_secret=check_env.check("CONTROLSHIFT_CLIENT_SECRET", client_secret),
            token_url=token_url,
            auto_refresh_url=token_url,
        )

    def get_petitions(self) -> Table:
        """
        Get a full list of all petitions, including ones that are unlaunched or otherwise not
        visible to the public.

        `Return:`
            Table Class
        """
        next_page = 1
        petitions = []
        while next_page:
            response = self.client.get_request(
                f"{self.hostname}/api/v1/petitions", {"page": next_page}
            )
            next_page = response["meta"]["next_page"]
            petitions.extend(response["petitions"])

        return Table(petitions)

        def get_signatures(self):
            pass

        def get_members(self):
            pass

        def get_partnerships(self):
            pass

        def get_events(self):
            pass

        def get_attendees(self):
            pass

        def get_calendars(self):
            pass

        def get_local_groups(self):
            pass



"""NGPVAN Score Endpoints"""

from parsons.etl.table import Table
from parsons.utilities import cloud_storage
import uuid
import logging
import petl

logger = logging.getLogger(__name__)


class Scores(object):
    def __init__(self, van_connection):

        self.connection = van_connection

    def get_scores(self):
        """
        Get all scores.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.connection.get_request("scores"))
        logger.info(f"Found {tbl.num_rows} scores.")
        return tbl

    def get_score(self, score_id):
        """
        Get an individual score.

        `Args:`
            score_id: int
                The score id
        `Returns:`
            dict
        """

        r = self.connection.get_request(f"scores/{score_id}")
        logger.info(f"Found score {score_id}.")
        return r

    def get_score_updates(self, created_before=None, created_after=None, score_id=None):
        """
        Get score updates.

        `Args:`
            created_before: str
                Filter score updates to those created before date. Use "YYYY-MM-DD"
                format.
            created_after: str
                Filter score updates to those created after date. Use "YYYY-MM-DD"
                format.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {
            "createdBefore": created_before,
            "createdAfter": created_after,
            "scoreId": score_id,
        }

        tbl = Table(self.connection.get_request("scoreUpdates", params=params))
        if tbl.num_rows:
            tbl.unpack_dict("updateStatistics", prepend=False)
            tbl.unpack_dict("score", prepend=False)
        logger.info(f"Found {tbl.num_rows} score updates.")
        return tbl

    def get_score_update(self, score_update_id):
        """
        Get a score update object

            `Args:`
                score_update_id : int
                        The score update id
            `Returns:`
                dict
        """

        r = self.connection.get_request(f"scoreUpdates/{score_update_id}")
        logger.info(f"Returning score update {score_update_id}.")
        return r

    def update_score_status(self, score_update_id, status):
        """
        Change the status of a score update object. This end point is used to
        approve a score loading job.

        `Args:`
            score_update_id: str
                The score update id
            status: str
                One of 'pending approval', 'approved', 'disapproved'
        `Returns:`
            ``None``
        """

        if status not in ["pending approval", "approved", "disapproved", "canceled"]:

            raise ValueError(
                """Valid inputs for status are, 'pending approval',
                             'approved','disapproved','canceled'"""
            )

        else:
            if status == "pending approval":
                status = "PendingApproval"
            else:
                status = status.capitalize()

        json = {"loadStatus": status}

        r = self.connection.patch_request(f"scoreUpdates/{score_update_id}", json=json)
        logger.info(f"Score {score_update_id} status updated to {status}.")
        return r

    def upload_scores(
        self,
        tbl,
        config,
        url_type,
        id_type="vanid",
        email=None,
        auto_approve=True,
        approve_tolerance=0.1,
        **url_kwargs,
    ):
        """
        Upload scores. Use to create or overwrite scores. Multiple score loads
        should be configured in a single call. [1]_

        `Args:`
            tbl: object
                A parsons.Table object. The table must contain the scores and first column in the
                table must contain the primary key (e.g. vanid).
            config: list
                The score configuration. A list of dictionaries in which you specify the following

                .. list-table::
                    :widths: 20 80
                    :header-rows: 0

                    * - ``score_column``
                      - The name of the column where the score is housed.
                    * - ``score_id``
                      - The score slot id.

                Example:

                .. highlight:: python
                .. code-block:: python

                  [{'score1_id' : int, score1_column': str}
                   {'score2_id' : int, score2_column': str}]

            url_type: str
                The cloud file storage to use to post the file (``S3`` or ``GCS``).
                See :ref:`Cloud Storage <cloud-storage>` for more details.
            email: str
                An email address to send job load status updates.
            auto_approve: boolean
                If the scores are within the expected tolerance of deviation from the
                average values provided, then score will be automatically approved.
            approve_tolderance: float
                The deviation from the average scores allowed in order to automatically
                approve the score. Maximum of .1.
            **url_kwargs: kwargs
                Arguments to configure your cloud storage url type. See
                :ref:`Cloud Storage <cloud-storage>` for more details.
        `Returns:`
            int
               The score load job id.

        .. [1] NGPVAN asks that you load multiple scores in a single call to reduce the load
           on their servers.
        """

        # Move to cloud storage
        file_name = str(uuid.uuid1())
        url = cloud_storage.post_file(
            tbl, url_type, file_path=file_name + ".zip", **url_kwargs
        )
        logger.info(f"Table uploaded to {url_type}.")

        # Generate shell request
        json = {
            "description": "A description",
            "file": {
                "columnDelimiter": "csv",
                "columns": [{"name": c} for c in tbl.columns],
                "fileName": file_name + ".csv",
                "hasHeader": "True",
                "hasQuotes": "False",
                "sourceUrl": url,
            },
            "actions": [],
        }

        # Configure each score
        for i in config:
            action = {
                "actionType": "score",
                "personIdColumn": tbl.columns[0],
                "personIdType": id_type,
                "scoreColumn": i["score_column"],
                "scoreId": i["score_id"],
            }

            if auto_approve:
                average = petl.stats(tbl.table, i["score_column"]).mean
                action["approvalCriteria"] = {
                    "average": average,
                    "tolerance": approve_tolerance,
                }

            json["actions"].append(action)

        # Add email listener
        if email:
            json["listeners"] = [{"type": "EMAIL", "value": email}]

        # Upload scores
        r = self.connection.post_request("fileLoadingJobs", json=json)
        logger.info(f"Scores job {r['jobId']} created.")
        return r["jobId"]


class FileLoadingJobs(object):
    def __init__(self, van_connection):

        self.connection = van_connection

    def create_file_load(
        self,
        file_name,
        file_url,
        columns,
        id_column,
        id_type,
        score_id,
        score_column,
        delimiter="csv",
        header=True,
        quotes=True,
        description=None,
        email=None,
        auto_average=None,
        auto_tolerance=None,
    ):
        """
        .. warning::
           .. deprecated:: 0.7 Use :func:`parsons.VAN.upload_scores` instead.

        Loads a file. Only used for loading scores at this time. Scores must be
        compressed using `zip`.

        `Args:`
            file_name: str
                The name of the file contained in the zip file.
            file_url: str
                The url path to directly download the file. Can also be a path to an FTP site.
            columns: list
                A list of column names contained in the file.
            id_column: str
                The column name of the id column in the file.
            id_type: str
                A valid primary key, such as `VANID` or `DWID`. Varies by VAN instance.
            score_id: int
                The score slot id
            score_column: str
                    The column holding the score
            delimiter: str
                    The file delimiter used.
            email: str
                A valid email address in which file loading status will be sent.
            auto_average: float
                The average of scores to be loaded.
            auto_tolerance: float
                The fault tolerance of the VAN calculated average compared to the ``auto_average``.
                The tolerance must be less than 10% of the difference between the maximum and
                minimum possible acceptable values of the score.
        `Returns:`
            dict
                The file load id
        """

        columns = [{"name": c} for c in columns]

        # To Do: Validate that it is a .zip file. Not entirely sure if this is possible
        # as some urls might not end in ".zip".

        if delimiter not in ["csv", "tab", "pipe"]:
            raise ValueError("Delimiter must be one of 'csv', 'tab' or 'pipe'")

        delimiter = delimiter.capitalize()

        json = {
            "description": "A description",
            "file": {
                "columnDelimiter": delimiter,
                "columns": columns,
                "fileName": file_name,
                "hasHeader": header,
                "hasQuotes": quotes,
                "sourceUrl": file_url,
            },
            "actions": [
                {
                    "actionType": "score",
                    "personIdColumn": id_column,
                    "personIdType": id_type,
                    "scoreColumn": score_column,
                    "scoreId": score_id,
                }
            ],
            "listeners": [{"type": "EMAIL", "value": email}],
        }

        if auto_average and auto_tolerance:

            json["actions"]["approvalCriteria"] = {
                "average": auto_average,
                "tolerance": auto_tolerance,
            }

        r = self.connection.post_request("fileLoadingJobs", json=json)["jobId"]
        logger.info(f"Score loading job {r} created.")
        return r

    def create_file_load_multi(
        self,
        file_name,
        file_url,
        columns,
        id_column,
        id_type,
        score_map,
        delimiter="csv",
        header=True,
        quotes=True,
        description=None,
        email=None,
    ):
        """
        .. warning::
           .. deprecated:: 0.7 Use :func:`parsons.VAN.upload_scores` instead.

        An iteration of the :meth:`file_load` method that allows you to load multiple scores
        at the same time.

        `Args:`
            file_name : str
                The name of the file contained in the zip file.
            file_url : str
                The url path to directly download the file. Can also be a path to an FTP site.
            columns: list
                A list of column names contained in the file.
            id_column : str
                The column name of the id column in the file.
            id_type : str
                A valid primary key, such as `VANID` or `DWID`. Varies by VAN instance.
            score_map : list
                A list of dicts that adheres to the following syntax

                .. highlight:: python
                .. code-block:: python

                    [{'score_id' : int,
                      'score_column': str,
                      'auto_average': float,
                      'auto_tolerance': float }]

            email: str
                A valid email address in which file loading status will be sent.
        `Returns:`
            The file load job id
        """

        columns = [{"name": c} for c in columns]

        # To Do: Validate that it is a .zip file. Not entirely sure if this is possible
        # as some urls might not end in ".zip".

        if delimiter not in ["csv", "tab", "pipe"]:
            raise ValueError("Delimiter must be one of 'csv', 'tab' or 'pipe'")

        delimiter = delimiter.capitalize()

        json = {
            "description": "A description",
            "file": {
                "columnDelimiter": delimiter,
                "columns": columns,
                "fileName": file_name,
                "hasHeader": header,
                "hasQuotes": quotes,
                "sourceUrl": file_url,
            },
            "listeners": [{"type": "EMAIL", "value": email}],
        }

        actions = []

        for score in score_map:

            action = {
                "actionType": "score",
                "personIdColumn": id_column,
                "personIdType": id_type,
                "scoreColumn": score["score_column"],
                "scoreId": score["score_id"],
                "approvalCriteria": {
                    "average": score["auto_average"],
                    "tolerance": score["auto_tolerance"],
                },
            }

            actions.append(action)

        json["actions"] = actions

        r = self.connection.post_request("fileLoadingJobs", json=json)["jobId"]
        logger.info(f"Score loading job {r} created.")
        return r



from dateutil.parser import parse
import datetime


def date_to_timestamp(value, tzinfo=datetime.timezone.utc):
    """Convert any date value into a Unix timestamp.

    `Args:`
        value: int or str or datetime
            Value to parse
        tzinfo: datetime.timezone
            `Optional`: Timezone for the datetime; defaults to UTC.
    `Returns:`
        Unix timestamp (int)
    """

    parsed_date = parse_date(value)

    if not parsed_date:
        return None

    if not parsed_date.tzinfo:
        parsed_date = parsed_date.replace(tzinfo=tzinfo)

    return int(parsed_date.timestamp())


def parse_date(value, tzinfo=datetime.timezone.utc):
    """Parse an arbitrary date value into a Python datetime.

    If no value is provided (i.e., the value is None or empty), then the return value will be
    None.

    `Args:`
        value: int or str or datetime
            Value to parse
        tzinfo: datetime.timezone
            `Optional`: Timezone for the datetime; defaults to UTC.
    `Returns:`
        datetime.datetime or None
    """

    if not value:
        return None

    # If it's a number, we (probably) have a unix timestamp
    if isinstance(value, int):
        parsed = datetime.datetime.fromtimestamp(value, tzinfo)
    elif isinstance(value, datetime.datetime):
        parsed = value
    elif isinstance(value, str):
        parsed = parse(value)
    else:
        raise TypeError(
            "Unable to parse value; must be one of string or int or datetime, but got type "
            f"{type(value)}"
        )

    if not parsed.tzinfo:
        parsed = parsed.replace(tzinfo=tzinfo)

    return parsed



class Questions:
    """A class to access the Questions PDI API endpoint."""

    def __init__(self):
        questions_endpoint = "/questions"
        self.url_questions = self.base_url + questions_endpoint

        super().__init__()

    def get_questions(self, limit=None):
        """Get a list of Questions.

        `Args:`
            limit: int
                Specify limit to return.

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        return self._request(self.url_questions, limit=limit)

    def get_question(self, id: str):
        """
        Get a Question by id.

        `Args:`
            id: str
                The Question id
        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        return self._request(f"{self.url_questions}/{id}")

    def create_question(
        self,
        question: str,
        type: str,
        category: str,
        answer_options: list,
        question_label: str = None,
        question_description: str = None,
        candidate_issue_id: str = None,
        default: bool = True,
        *args,
    ):
        """
        answer_options:[
                {
                "id": "string",
                "flagId": "string",
                "displayDescription": "string",
                "displayCode": "string"
                }
            ]
        """
        payload = {
            "question": question,
            "questionLabel": question_label,
            "questionDescription": question_description,
            "type": type,
            "category": category,
            "candidateIssueId": candidate_issue_id,
            "default": default,
            "answerOptions": answer_options,
        }
        return self._request(self.locations_url, req_type="POST", post_data=payload)

    def delete_question(self, id: str):
        return self._request(f"{self.url_questions}/{id}", req_type="DELETE")



import os
import unittest
import requests_mock
from parsons import VAN, Table
from test.utils import validate_list, assert_matching_tables
from requests.exceptions import HTTPError


class TestNGPVAN(unittest.TestCase):
    def setUp(self):

        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_canvass_responses_contact_types(self, m):

        json = {"name": "Auto Dial", "contactTypeId": 19, "channelTypeName": "Phone"}

        m.get(self.van.connection.uri + "canvassResponses/contactTypes", json=json)

        assert_matching_tables(
            Table(json), self.van.get_canvass_responses_contact_types()
        )

    @requests_mock.Mocker()
    def test_get_canvass_responses_input_types(self, m):

        json = {"inputTypeId": 11, "name": "API"}
        m.get(self.van.connection.uri + "canvassResponses/inputTypes", json=json)
        assert_matching_tables(
            Table(json), self.van.get_canvass_responses_input_types()
        )

    @requests_mock.Mocker()
    def test_get_canvass_responses_result_codes(self, m):

        json = {
            "shortName": "BZ",
            "resultCodeId": 18,
            "name": "Busy",
            "mediumName": "Busy",
        }

        m.get(self.van.connection.uri + "canvassResponses/resultCodes", json=json)
        assert_matching_tables(
            Table(json), self.van.get_canvass_responses_result_codes()
        )

    @requests_mock.Mocker()
    def test_get_survey_questions(self, m):

        json = {
            "count": 67,
            "items": [
                {
                    "status": "Active",
                    "responses": [
                        {
                            "shortName": "1",
                            "surveyResponseId": 1288926,
                            "name": "1-Strong Walz",
                            "mediumName": "1",
                        },
                        {
                            "shortName": "2",
                            "surveyResponseId": 1288928,
                            "name": "2-Lean Walz",
                            "mediumName": "2",
                        },
                    ],
                    "scriptQuestion": "Who do you support for Governor?",
                    "name": "MN Governor Gen",
                    "surveyQuestionId": 311838,
                    "mediumName": "MNGovG",
                    "shortName": "MGG",
                    "type": "Candidate",
                    "cycle": 2018,
                }
            ],
            "nextPageLink": None,
        }

        m.get(self.van.connection.uri + "surveyQuestions", json=json)

        expected = [
            "status",
            "responses",
            "scriptQuestion",
            "name",
            "surveyQuestionId",
            "mediumName",
            "shortName",
            "type",
            "cycle",
        ]

        self.assertTrue(validate_list(expected, self.van.get_survey_questions()))

    @requests_mock.Mocker()
    def test_get_supporter_groups(self, m):

        json = {
            "items": [
                {"id": 12, "name": "tmc", "description": "A fun group."},
                {"id": 13, "name": "tmc", "description": "A fun group."},
            ],
            "nextPageLink": None,
            "count": 3,
        }

        m.get(self.van.connection.uri + "supporterGroups", json=json)

        ["id", "name", "description"]

        self.van.get_supporter_groups()

    @requests_mock.Mocker()
    def test_get_supporter_group(self, m):

        json = {"id": 12, "name": "tmc", "description": "A fun group."}
        m.get(self.van.connection.uri + "supporterGroups/12", json=json)

        # Test that columns are expected columns
        self.assertEqual(self.van.get_supporter_group(12), json)

    @requests_mock.Mocker()
    def test_delete_supporter_group(self, m):

        # Test good input
        good_supporter_group_id = 5
        good_ep = f"supporterGroups/{good_supporter_group_id}"
        m.delete(self.van.connection.uri + good_ep, status_code=204)
        self.van.delete_supporter_group(good_supporter_group_id)

        # Test bad input raises
        bad_supporter_group_id = 999
        # bad_vanid = 99999
        bad_ep = f"supporterGroups/{bad_supporter_group_id}"
        m.delete(self.van.connection.uri + bad_ep, status_code=404)
        self.assertRaises(
            HTTPError, self.van.delete_supporter_group, bad_supporter_group_id
        )

    @requests_mock.Mocker()
    def test_add_person_supporter_group(self, m):

        # Test good input
        good_supporter_group_id = 5
        good_vanid = 12345
        good_uri = f"supporterGroups/{good_vanid}/people/{good_supporter_group_id}"
        m.put(self.van.connection.uri + good_uri, status_code=204)
        self.van.add_person_supporter_group(good_vanid, good_supporter_group_id)

        # Test bad input
        bad_supporter_group_id = 999
        bad_vanid = 99999
        bad_uri = f"supporterGroups/{bad_vanid}/people/{bad_supporter_group_id}"
        m.put(self.van.connection.uri + bad_uri, status_code=404)
        self.assertRaises(
            HTTPError,
            self.van.add_person_supporter_group,
            bad_vanid,
            bad_supporter_group_id,
        )

    @requests_mock.Mocker()
    def test_delete_person_supporter_group(self, m):

        # Test good input
        good_supporter_group_id = 5
        good_vanid = 12345
        good_ep = f"supporterGroups/{good_vanid}/people/{good_supporter_group_id}"
        m.delete(self.van.connection.uri + good_ep, status_code=204)
        self.van.delete_person_supporter_group(good_vanid, good_supporter_group_id)

        # Test bad input raises
        bad_supporter_group_id = 999
        bad_vanid = 99999
        bad_ep = f"supporterGroups/{bad_vanid}/people/{bad_supporter_group_id}"
        m.delete(self.van.connection.uri + bad_ep, status_code=404)
        self.assertRaises(
            HTTPError,
            self.van.delete_person_supporter_group,
            bad_vanid,
            bad_supporter_group_id,
        )


if __name__ == "__main__":

    unittest.main()



import csv
import io
import gzip

import petl
import pytest
from parsons.targetsmart.targetsmart_api import TargetSmartAPI


@pytest.fixture
def intable():
    return petl.wrap(
        [
            [
                "first_name",
                "last_name",
                "address1",
                "email",
                "phone",
                "some_unknown_field",
            ],
            ["Bob", "Smith", "123 Main", "bob@example.com", "1231231234", "foo"],
            ["Alice", "Example", "123 Main", "", "", "bar"],
            ["Sally", "Example", "123 Main", "", "", ""],
        ]
    )


@pytest.fixture
def raw_outtable(intable):
    return (
        intable.addrownumbers(field="ts__input_row")
        .addrownumbers(field="ts__row")
        .addrownumbers(field="matchback_id")
        .convert("ts__input_row", str)
        .convert("ts__row", str)
        .addfield("tsmart_match_code", "Y")
        .addfield("vb.voterbase_id", "OH-123")
    )


@pytest.fixture
def prep_intable(intable):
    return intable.addrownumbers(field="matchback_id")


@pytest.fixture
def raw_outcsv(raw_outtable):
    buf = io.StringIO()
    writer = csv.writer(buf)
    writer.writerows(list(raw_outtable))
    return buf.getvalue()


@pytest.fixture
def raw_outgz(raw_outcsv):
    buf = io.BytesIO()
    with gzip.GzipFile(fileobj=buf, mode="w") as gz:
        gz.write(raw_outcsv.encode("utf8"))
    return buf.getvalue()


@pytest.fixture
def final_outtable(prep_intable, raw_outtable):
    return petl.leftjoin(prep_intable, raw_outtable, key="matchback_id").cutout(
        "matchback_id"
    )


@pytest.fixture
def submit_filename():
    return "parsons_test.csv"


def test_smartmatch(
    intable,
    submit_filename,
    raw_outgz,
    raw_outcsv,
    raw_outtable,
    final_outtable,
    requests_mock,
):
    ts = TargetSmartAPI("mockkey")
    resp1 = {"url": "https://mock_smartmatch_upload_endpoint", "error": None}
    poll_resp = {"url": "https://mock_smartmatch_download_endpoint", "error": None}
    requests_mock.get("https://api.targetsmart.com/service/smartmatch", json=resp1)
    requests_mock.put(resp1["url"])
    requests_mock.get(
        "https://api.targetsmart.com/service/smartmatch/poll", json=poll_resp
    )
    requests_mock.get(poll_resp["url"], content=raw_outgz)

    results = ts.smartmatch(intable).to_petl()
    assert list(final_outtable) == list(results)



import json

import requests
from parsons.etl.table import Table
from parsons.utilities import check_env


class Auth0(object):
    """
    Instantiate the Auth0 class

    `Args:`
        client_id: str
            The Auth0 client ID. Not required if ``AUTH0_CLIENT_ID`` env variable set.
        client_secret: str
            The Auth0 client secret. Not required if ``AUTH0_CLIENT_SECRET`` env variable set.
        domain: str
            The Auth0 domain. Not required if ``AUTH0_DOMAIN`` env variable set.
    `Returns:`
        Auth0 Class
    """

    def __init__(self, client_id=None, client_secret=None, domain=None):
        self.base_url = f"https://{check_env.check('AUTH0_DOMAIN', domain)}"
        access_token = (
            requests.post(
                f"{self.base_url}/oauth/token",
                data={
                    "grant_type": "client_credentials",  # OAuth 2.0 flow to use
                    "client_id": check_env.check("AUTH0_CLIENT_ID", client_id),
                    "client_secret": check_env.check(
                        "AUTH0_CLIENT_SECRET", client_secret
                    ),
                    "audience": f"{self.base_url}/api/v2/",
                },
            )
            .json()
            .get("access_token")
        )
        self.headers = {
            "Authorization": f"Bearer {access_token}",
            "Content-Type": "application/json",
        }

    def delete_user(self, id):
        """
        Delete Auth0 user.

        `Args:`
            id: str
                The user ID of the record to delete.
        `Returns:`
            int
        """
        return requests.delete(
            f"{self.base_url}/api/v2/users/{id}", headers=self.headers
        ).status_code

    def get_users_by_email(self, email):
        """
        Get Auth0 users by email.

        `Args:`
            email: str
                The user email of the record to get.
        `Returns:`
            Table Class
        """
        return Table(
            requests.get(
                f"{self.base_url}/api/v2/users-by-email",
                headers=self.headers,
                params={"email": email},
            ).json()
        )

    def upsert_user(
        self,
        email,
        username=None,
        given_name=None,
        family_name=None,
        app_metadata={},
        user_metadata={},
    ):
        """
        Upsert Auth0 users by email.

        `Args:`
            email: str
                The user email of the record to get.
            username: optional str
                Username to set for user
            given_name: optional str
                Given to set for user
            family_name: optional str
                Family name to set for user
            app_metadata: optional dict
                App metadata to set for user
            user_metadata: optional dict
                User metadata to set for user
        `Returns:`
            Requests Response object
        """
        payload = json.dumps(
            {
                "email": email.lower(),
                "given_name": given_name,
                "family_name": family_name,
                "username": username,
                "connection": "Username-Password-Authentication",
                "app_metadata": app_metadata,
                "blocked": False,
                "user_metadata": user_metadata,
            }
        )
        existing = self.get_users_by_email(email.lower())
        if existing.num_rows > 0:
            a0id = existing[0]["user_id"]
            ret = requests.patch(
                f"{self.base_url}/api/v2/users/{a0id}",
                headers=self.headers,
                data=payload,
            )
        else:
            ret = requests.post(
                f"{self.base_url}/api/v2/users", headers=self.headers, data=payload
            )
        if ret.status_code != 200:
            raise ValueError(f"Invalid response {ret.json()}")
        return ret



from parsons.bluelink.bluelink import Bluelink
from parsons.bluelink.person import (
    BluelinkPerson,
    BluelinkEmail,
    BluelinkAddress,
    BluelinkPhone,
    BluelinkIdentifier,
    BluelinkTag,
    BluelinkScore,
)

__all__ = [
    "Bluelink",
    "BluelinkPerson",
    "BluelinkEmail",
    "BluelinkAddress",
    "BluelinkPhone",
    "BluelinkIdentifier",
    "BluelinkTag",
    "BluelinkScore",
]



import os
import zipfile
from parsons.utilities.files import create_temp_directory


def create_archive(archive_path, file_path, file_name=None, if_exists="replace"):
    """
    Create and fill an archive.

    `Args:`
        archive_path: str
            The file name of zip archive
        file_path: str
            The path of the file
        file_name: str
            The name of the file in the archive
        if_exists: str
            If archive already exists, one of 'replace' or 'append'
    `Returns:`
        Zip archive path
    """

    if if_exists == "append":
        write_type = "a"
    else:
        write_type = "w"

    if not file_name:
        file_name = file_path.split("/")[-1]

    with zipfile.ZipFile(archive_path, write_type) as z:
        z.write(file_path, arcname=file_name, compress_type=zipfile.ZIP_STORED)

    return archive_path


def unzip_archive(archive_path, destination=None):
    """
    Unzip an archive. Only returns the path of the first
    file in the archive.

    `Args:`
        archive_path: str
            Path to the ZIP archive
        destination: str
            `Optional`; path to unzip the archive into; if not specified, the
    `Returns:`
        Extracted file path.
    """
    destination = destination or create_temp_directory()

    with zipfile.ZipFile(archive_path, "r") as z:
        file_name = z.namelist()[0]
        z.extractall(path=destination)
        return os.path.join(destination, file_name)



import json
import logging
import re
import warnings

from parsons import Table
from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector

logger = logging.getLogger(__name__)

API_URL = "https://actionnetwork.org/api/v2"


class ActionNetwork(object):
    """
    `Args:`
        api_token: str
            The OSDI API token
    """

    def __init__(self, api_token=None):
        self.api_token = check_env.check("AN_API_TOKEN", api_token)
        self.headers = {
            "Content-Type": "application/json",
            "OSDI-API-Token": self.api_token,
        }
        self.api_url = API_URL
        self.api = APIConnector(self.api_url, headers=self.headers)

    def _get_page(self, object_name, page, per_page=25, filter=None):
        # returns data from one page of results
        if per_page > 25:
            per_page = 25
            logger.info(
                "Action Network's API will not return more than 25 entries per page. \
            Changing per_page parameter to 25."
            )
        params = {"page": page, "per_page": per_page, "filter": filter}
        return self.api.get_request(url=object_name, params=params)

    def _get_entry_list(self, object_name, limit=None, per_page=25, filter=None):
        # returns a list of entries for a given object, such as people, tags, or actions
        # Filter can only be applied to people, petitions, events, forms, fundraising_pages,
        # event_campaigns, campaigns, advocacy_campaigns, signatures, attendances, submissions,
        # donations and outreaches.
        # See Action Network API docs for more info: https://actionnetwork.org/docs/v2/
        count = 0
        page = 1
        return_list = []
        while True:
            response = self._get_page(object_name, page, per_page, filter=filter)
            page = page + 1
            response_list = response["_embedded"][f"osdi:{object_name}"]
            if not response_list:
                return Table(return_list)
            return_list.extend(response_list)
            count = count + len(response_list)
            if limit:
                if count >= limit:
                    return Table(return_list[0:limit])

    def get_people(self, limit=None, per_page=25, page=None, filter=None):
        """
        `Args:`
            limit:
                The number of entries to return. When None, returns all entries.
            per_page
                The number of entries per page to return. 25 maximum.
            page
                Which page of results to return
            filter
                The OData query for filtering results. E.g. "modified_date gt '2014-03-25'".
                When None, no filter is applied.
        `Returns:`
            A list of JSONs of people stored in Action Network.
        """
        if page:
            return self._get_page("people", page, per_page, filter=filter)
        return self._get_entry_list("people", limit, per_page, filter=filter)

    def get_person(self, person_id):
        """
        `Args:`
            person_id:
                Id of the person.
        `Returns:`
            A  JSON of the entry. If the entry doesn't exist, Action Network returns
            ``{'error': 'Couldn't find person with id = <id>'}``.
        """
        return self.api.get_request(url=f"people/{person_id}")

    def upsert_person(
        self,
        email_address=None,
        given_name=None,
        family_name=None,
        tags=None,
        languages_spoken=None,
        postal_addresses=None,
        mobile_number=None,
        mobile_status="subscribed",
        background_processing=False,
        identifiers=None,
        **kwargs,
    ):
        """
        Creates or updates a person record. In order to update an existing record instead of
        creating a new one, you must supply an email or mobile number which matches a record
        in the database.

        `Args:`
            email_address:
                Either email_address or mobile_number are required. Can be any of the following
                    - a string with the person's email
                    - a list of strings with a person's emails
                    - a dictionary with the following fields
                        - email_address (REQUIRED)
                        - primary (OPTIONAL): Boolean indicating the user's primary email address
                        - status (OPTIONAL): can taken on any of these values
                            - "subscribed"
                            - "unsubscribed"
                            - "bouncing"
                            - "previous bounce"
                            - "spam complaint"
                            - "previous spam complaint"
            given_name:
                The person's given name
            family_name:
                The person's family name
            tags:
                Optional field. A list of strings of pre-existing tags to be applied to the person.
            languages_spoken:
                Optional field. A list of strings of the languages spoken by the person
            postal_addresses:
                Optional field. A list of dictionaries.
                For details, see Action Network's documentation:
                https://actionnetwork.org/docs/v2/person_signup_helper
            mobile_number:
                Either email_address or mobile_number are required. Can be any of the following
                    - a string with the person's cell phone number
                    - an integer with the person's cell phone number
                    - a list of strings with the person's cell phone numbers
                    - a list of integers with the person's cell phone numbers
                    - a dictionary with the following fields
                        - number (REQUIRED)
                        - primary (OPTIONAL): Boolean indicating the user's primary mobile number
                        - status (OPTIONAL): can taken on any of these values
                            - "subscribed"
                            - "unsubscribed"
            mobile_status:
                'subscribed' or 'unsubscribed'
            background_request: bool
                If set `true`, utilize ActionNetwork's "background processing". This will return
                an immediate success, with an empty JSON body, and send your request to the
                background queue for eventual processing.
                https://actionnetwork.org/docs/v2/#background-processing
            identifiers:
                List of strings to be used as globally unique
                identifiers. Can be useful for matching contacts back
                to other platforms and systems. If the identifier
                provided is not globally unique in ActionNetwork, it will
                simply be ignored and not added to the object. Action Network
                also creates its own identifier for each new resouce.
                https://actionnetwork.org/docs/v2/#resources
                e.g.: ["foreign_system:1", "other_system:12345abcd"]

            **kwargs:
                Any additional fields to store about the person. Action Network allows
                any custom field.
        Adds a person to Action Network
        """
        email_addresses_field = None
        if type(email_address) == str:
            email_addresses_field = [{"address": email_address}]
        elif type(email_address) == list:
            if type(email_address[0]) == str:
                email_addresses_field = [{"address": email} for email in email_address]
                email_addresses_field[0]["primary"] = True
            if type(email_address[0]) == dict:
                email_addresses_field = email_address

        mobile_numbers_field = None
        if type(mobile_number) == str:
            mobile_numbers_field = [
                {"number": re.sub("[^0-9]", "", mobile_number), "status": mobile_status}
            ]
        elif type(mobile_number) == int:
            mobile_numbers_field = [
                {"number": str(mobile_number), "status": mobile_status}
            ]
        elif type(mobile_number) == list:
            if len(mobile_number) > 1:
                raise ("Action Network allows only 1 phone number per activist")
            if type(mobile_number[0]) == str:
                mobile_numbers_field = [
                    {"number": re.sub("[^0-9]", "", cell), "status": mobile_status}
                    for cell in mobile_number
                ]
                mobile_numbers_field[0]["primary"] = True
            if type(mobile_number[0]) == int:
                mobile_numbers_field = [
                    {"number": cell, "status": mobile_status} for cell in mobile_number
                ]
                mobile_numbers_field[0]["primary"] = True
            if type(mobile_number[0]) == dict:
                mobile_numbers_field = mobile_number

        if not email_addresses_field and not mobile_numbers_field:
            raise (
                "Either email_address or mobile_number is required and can be formatted "
                "as a string, list of strings, a dictionary, a list of dictionaries, or "
                "(for mobile_number only) an integer or list of integers"
            )

        data = {"person": {}}

        if email_addresses_field is not None:
            data["person"]["email_addresses"] = email_addresses_field
        if mobile_numbers_field is not None:
            data["person"]["phone_numbers"] = mobile_numbers_field
        if given_name is not None:
            data["person"]["given_name"] = given_name
        if family_name is not None:
            data["person"]["family_name"] = family_name
        if languages_spoken is not None:
            data["person"]["languages_spoken"] = languages_spoken
        if postal_addresses is not None:
            data["person"]["postal_addresses"] = postal_addresses
        if tags is not None:
            data["add_tags"] = tags
        if identifiers:
            data["person"]["identifiers"] = identifiers
        data["person"]["custom_fields"] = {**kwargs}
        url = f"{self.api_url}/people"
        if background_processing:
            url = f"{url}?background_processing=true"
        response = self.api.post_request(url, data=json.dumps(data))

        identifiers = response["identifiers"]
        person_id = [
            entry_id.split(":")[1]
            for entry_id in identifiers
            if "action_network:" in entry_id
        ][0]
        if response["created_date"] == response["modified_date"]:
            logger.info(f"Entry {person_id} successfully added.")
        else:
            logger.info(f"Entry {person_id} successfully updated.")
        return response

    def add_person(
        self,
        email_address=None,
        given_name=None,
        family_name=None,
        tags=None,
        languages_spoken=None,
        postal_addresses=None,
        mobile_number=None,
        mobile_status="subscribed",
        **kwargs,
    ):
        """
        Creates a person in the database. WARNING: this endpoint has been deprecated in favor of
        upsert_person.
        """
        logger.warning(
            "Method 'add_person' has been deprecated. Please use 'upsert_person'."
        )
        # Pass inputs to preferred method:
        self.upsert_person(
            email_address=email_address,
            given_name=given_name,
            family_name=family_name,
            languages_spoken=languages_spoken,
            postal_addresses=postal_addresses,
            mobile_number=mobile_number,
            mobile_status=mobile_status,
            **kwargs,
        )

    def update_person(self, entry_id, background_processing=False, **kwargs):
        """
        Updates a person's data in Action Network, given their Action Network ID. Note that you
        can't alter a person's tags with this method. Instead, use upsert_person.

        `Args:`
            entry_id:
                The person's Action Network id
            background_processing: bool
                If set `true`, utilize ActionNetwork's "background processing". This will return
                an immediate success, with an empty JSON body, and send your request to the
                background queue for eventual processing.
                https://actionnetwork.org/docs/v2/#background-processing
            **kwargs:
                Fields to be updated. The possible fields are
                    email_address:
                        Can be any of the following
                            - a string with the person's email
                            - a dictionary with the following fields
                                - email_address (REQUIRED)
                                    - primary (OPTIONAL): Boolean indicating the user's
                                    primary email address
                                - status (OPTIONAL): can taken on any of these values
                                    - "subscribed"
                                    - "unsubscribed"
                                    - "bouncing"
                                    - "previous bounce"
                                    - "spam complaint"
                                    - "previous spam complaint"
                    given_name:
                        The person's given name
                    family_name:
                        The person's family name
                    languages_spoken:
                        Optional field. A list of strings of the languages spoken by the person
                    postal_addresses:
                        Optional field. A list of dictionaries.
                        For details, see Action Network's documentation:
                        https://actionnetwork.org/docs/v2/people#put
                    custom_fields:
                        A dictionary of any other fields to store about the person.
        """
        data = {**kwargs}
        url = f"{self.api_url}/people/{entry_id}"
        if background_processing:
            url = f"{url}?background_processing=true"
        response = self.api.put_request(
            url=url,
            data=json.dumps(data),
            success_codes=[204, 201, 200],
        )
        logger.info(f"Person {entry_id} successfully updated")
        return response

    def get_tags(self, limit=None, per_page=None):
        """
        `Args:`
            limit:
                The number of entries to return. When None, returns all entries.
            per_page:
                This is a deprecated argument.
        `Returns:`
            A list of JSONs of tags in Action Network.
        """
        if per_page:
            warnings.warn(
                "per_page is a deprecated argument on get_tags()",
                DeprecationWarning,
                stacklevel=2,
            )
        return self._get_entry_list("tags", limit)

    def get_tag(self, tag_id):
        """
        `Args:`
            tag_id:
                Id of the tag.
        `Returns:`
            A  JSON of the entry. If the entry doesn't exist, Action Network returns
            "{'error': 'Couldn't find tag with id = <id>'}"
        """
        return self.api.get_request(url=f"tags/{tag_id}")

    def add_tag(self, name):
        """
        `Args:`
            name:
                The tag's name. This is the ONLY editable field
        Adds a tag to Action Network. Once created, tags CANNOT be edited or deleted.
        """
        data = {"name": name}
        response = self.api.post_request(
            url=f"{self.api_url}/tags", data=json.dumps(data)
        )
        identifiers = response["identifiers"]
        person_id = [
            entry_id.split(":")[1]
            for entry_id in identifiers
            if "action_network:" in entry_id
        ][0]
        logger.info(f"Tag {person_id} successfully added to tags.")
        return response

    def create_event(self, title, start_date=None, location=None):
        """
        Create an event in Action Network

        `Args:`
            title: str
                The public title of the event
            start_date: str OR datetime
                OPTIONAL: The starting date & time. If a string, use format "YYYY-MM-DD HH:MM:SS"
                (hint: the default format you get when you use `str()` on a datetime)
            location: dict
                OPTIONAL: A dict of location details. Can include any combination of the types of
                values in the following example:
                .. code-block:: python

                    my_location = {
                        "venue": "White House",
                        "address_lines": [
                            "1600 Pennsylvania Ave"
                        ],
                        "locality": "Washington",
                        "region": "DC",
                        "postal_code": "20009",
                        "country": "US"
                    }

        `Returns:`
            Dict of Action Network Event data.
        """

        data = {"title": title}

        if start_date:
            start_date = str(start_date)
            data["start_date"] = start_date

        if isinstance(location, dict):
            data["location"] = location

        event_dict = self.api.post_request(
            url=f"{self.api_url}/events", data=json.dumps(data)
        )

        an_event_id = event_dict["_links"]["self"]["href"].split("/")[-1]
        event_dict["event_id"] = an_event_id

        return event_dict



import importlib
import inspect
import json
import os

import boto3


"""

In lambda handler:

from parsons.aws import event_command

def handler(event, context):

    ## ADD THESE TWO LINES TO TOP OF HANDLER:
    if event_command(event, context):
         return

"""
try:
    from zappa.asynchronous import run as zappa_run
except ImportError:
    zappa_run = None


def event_command(event, context):
    """
    Minimal `shim <https://en.wikipedia.org/wiki/Shim_(computing)>`_
    to add to the top lambda handler function to enable distributed tasks

    The rest of this library is compatible with zappa.async library.
    If you have deployed your app with `Zappa <https://github.com/Miserlou/Zappa>`_,
    then you do NOT need to add this shim.
    """
    if not set(event).intersection({"task_path", "args", "kwargs"}):
        return False  # did not match an event command
    func = import_and_get_task(event["task_path"], event.get("func_class_init_kwargs"))
    # if the func was decorated with zappa.async.task then run the real function
    func = getattr(func, "sync", func)

    # DID match an event command
    # -- so probably don't do the usual thing the Lambda handler does
    return func(*event["args"], **event["kwargs"]) or True


def run(
    func,
    args=[],
    kwargs={},
    service="lambda",
    capture_response=False,
    remote_aws_lambda_function_name=None,
    remote_aws_region=None,
    func_class=None,
    func_class_init_kwargs=None,
    **task_kwargs,
):
    lambda_function_name = remote_aws_lambda_function_name or os.environ.get(
        "AWS_LAMBDA_FUNCTION_NAME"
    )
    if not lambda_function_name or lambda_function_name == "FORCE_LOCAL":
        # We are neither running in Lambda environment, nor given one to invoke
        # so let's run it synchronously -- so code can be compatible both in-and-out of Lambda
        func(*args, **kwargs)
        return True
    # zappa has more robust and allows more configs -- but is not compatible with func_class
    if zappa_run and not func_class:
        return zappa_run(
            func,
            args,
            kwargs,
            service,
            capture_response,
            remote_aws_lambda_function_name,
            remote_aws_region,
            **task_kwargs,
        )

    task_path = get_func_task_path(func, func_class)
    payload = json.dumps(
        {
            "task_path": task_path,
            "args": args,
            "kwargs": kwargs,
            "func_class_init_kwargs": func_class_init_kwargs,
        }
    ).encode("utf-8")
    if len(payload) > 128000:  # pragma: no cover
        raise AsyncException("Payload too large for async Lambda call")
    lambda_client = boto3.Session().client("lambda")
    response = lambda_client.invoke(
        FunctionName=lambda_function_name,
        InvocationType="Event",  # makes the call async
        Payload=payload,
    )
    return response.get("StatusCode", 0) == 202


##
# Utility Functions
##


def import_and_get_task(task_path, instance_init_kwargs=None):
    """
    Given a modular path to a function, import that module
    and return the function.
    """
    module, function = task_path.rsplit(".", 1)
    app_module = importlib.import_module(module)
    class_func = function.split("|")
    app_function = getattr(app_module, class_func[0])
    if len(class_func) == 1:
        return app_function

    def init_and_run(*args, **kwargs):
        print("INITRUN", args, kwargs)
        if len(class_func) == 3:  # instance
            instance = app_function  # actually the class
        else:
            instance = app_function(**(instance_init_kwargs or {}))
        method = getattr(instance, class_func[1])
        return method(*args, **kwargs)

    return init_and_run


def get_func_task_path(func, method_class=None):
    """
    Format the modular task path for a function via inspection.
    """
    module_path = inspect.getmodule(method_class or func).__name__
    func_name = func.__name__

    # To support class methods, we need to see if it IS a method on a class
    # and then also determine if it is an instance method or a classmethod
    # Then we record that info with |'s to be decoded in import_and_get_task
    # classmethod format: "Foo|method|"
    # instance method format: "Foo|method"
    task_path = "{}.{}{}{}".format(
        module_path,
        f"{method_class.__name__}|" if method_class else "",
        func_name,
        "|" if method_class and "of <class" in repr(func) else "",
    )
    return task_path


class AsyncException(Exception):
    pass



GET_PEOPLE_RESPONSE = {
    "results": [
        {
            "birthdate": None,
            "city_district": None,
            "civicrm_id": None,
            "county_district": None,
            "county_file_id": None,
            "created_at": "2023-06-22T11:41:56-07:00",
            "datatrust_id": None,
            "do_not_call": False,
            "do_not_contact": False,
            "dw_id": None,
            "email": "foo@example.com",
            "email_opt_in": True,
            "employer": None,
            "external_id": None,
            "federal_district": None,
            "fire_district": None,
            "first_name": "Foo",
            "has_facebook": False,
            "id": 4,
            "is_twitter_follower": False,
            "is_volunteer": False,
            "judicial_district": None,
            "labour_region": None,
            "last_name": "Bar",
            "linkedin_id": None,
            "mobile": None,
            "mobile_opt_in": None,
            "middle_name": "",
            "nbec_guid": None,
            "ngp_id": None,
            "note": None,
            "occupation": None,
            "party": None,
            "pf_strat_id": None,
            "phone": None,
            "precinct_id": None,
            "primary_address": None,
            "profile_image_url_ssl": "https://example.com/assets/notifier/profile-avatar.png",
            "recruiter_id": None,
            "rnc_id": None,
            "rnc_regid": None,
            "salesforce_id": None,
            "school_district": None,
            "school_sub_district": None,
            "sex": None,
            "signup_type": 0,
            "state_file_id": None,
            "state_lower_district": None,
            "state_upper_district": None,
            "support_level": None,
            "supranational_district": None,
            "tags": ["zoot", "boot"],
            "twitter_id": None,
            "twitter_name": None,
            "updated_at": "2023-06-22T11:41:56-07:00",
            "van_id": None,
            "village_district": None,
            "ward": None,
            "work_phone_number": None,
        },
        {
            "birthdate": None,
            "city_district": None,
            "civicrm_id": None,
            "county_district": None,
            "county_file_id": None,
            "created_at": "2023-06-22T08:21:00-07:00",
            "datatrust_id": None,
            "do_not_call": False,
            "do_not_contact": False,
            "dw_id": None,
            "email": "bar@example.com",
            "email_opt_in": True,
            "employer": None,
            "external_id": None,
            "federal_district": None,
            "fire_district": None,
            "first_name": "Zoo",
            "has_facebook": False,
            "id": 2,
            "is_twitter_follower": False,
            "is_volunteer": False,
            "judicial_district": None,
            "labour_region": None,
            "last_name": "Baz",
            "linkedin_id": None,
            "mobile": None,
            "mobile_opt_in": True,
            "middle_name": "",
            "nbec_guid": None,
            "ngp_id": None,
            "note": None,
            "occupation": None,
            "party": None,
            "pf_strat_id": None,
            "phone": None,
            "precinct_id": None,
            "primary_address": None,
            "profile_image_url_ssl": "https://example.com/assets/notifier/profile-avatar.png",
            "recruiter_id": None,
            "rnc_id": None,
            "rnc_regid": None,
            "salesforce_id": None,
            "school_district": None,
            "school_sub_district": None,
            "sex": None,
            "signup_type": 0,
            "state_file_id": None,
            "state_lower_district": None,
            "state_upper_district": None,
            "support_level": None,
            "supranational_district": None,
            "tags": ["zoo", "bar"],
            "twitter_id": None,
            "twitter_name": None,
            "updated_at": "2023-06-22T11:43:03-07:00",
            "van_id": None,
            "village_district": None,
            "ward": None,
            "work_phone_number": None,
        },
    ],
    "next": None,
    "prev": None,
}

PERSON_RESPONSE = {
    "person": {
        "birthdate": None,
        "city_district": None,
        "civicrm_id": None,
        "county_district": None,
        "county_file_id": None,
        "created_at": "2023-06-22T08:21:00-07:00",
        "datatrust_id": None,
        "do_not_call": False,
        "do_not_contact": False,
        "dw_id": None,
        "email": "foo@example.com",
        "email_opt_in": True,
        "employer": None,
        "external_id": None,
        "federal_district": None,
        "fire_district": None,
        "first_name": "Foo",
        "has_facebook": False,
        "id": 1,
        "is_twitter_follower": False,
        "is_volunteer": False,
        "judicial_district": None,
        "labour_region": None,
        "last_name": "Bar",
        "linkedin_id": None,
        "mobile": None,
        "mobile_opt_in": True,
        "middle_name": "",
        "nbec_guid": None,
        "ngp_id": None,
        "note": None,
        "occupation": None,
        "party": None,
        "pf_strat_id": None,
        "phone": None,
        "precinct_id": None,
        "primary_address": None,
        "profile_image_url_ssl": "https://example.com/assets/notifier/profile-avatar.png",
        "recruiter_id": None,
        "rnc_id": None,
        "rnc_regid": None,
        "salesforce_id": None,
        "school_district": None,
        "school_sub_district": None,
        "sex": None,
        "signup_type": 0,
        "state_file_id": None,
        "state_lower_district": None,
        "state_upper_district": None,
        "support_level": None,
        "supranational_district": None,
        "tags": [],
        "twitter_id": None,
        "twitter_name": None,
        "updated_at": "2023-06-22T08:21:00-07:00",
        "van_id": None,
        "village_district": None,
        "ward": None,
        "work_phone_number": None,
        "active_customer_expires_at": None,
        "active_customer_started_at": None,
        "author": None,
        "author_id": None,
        "auto_import_id": None,
        "availability": None,
        "ballots": [],
        "banned_at": None,
        "billing_address": None,
        "bio": None,
        "call_status_id": None,
        "call_status_name": None,
        "capital_amount_in_cents": 500,
        "children_count": 0,
        "church": None,
        "city_sub_district": None,
        "closed_invoices_amount_in_cents": None,
        "closed_invoices_count": None,
        "contact_status_id": None,
        "contact_status_name": None,
        "could_vote_status": False,
        "demo": None,
        "donations_amount_in_cents": 0,
        "donations_amount_this_cycle_in_cents": 0,
        "donations_count": 0,
        "donations_count_this_cycle": 0,
        "donations_pledged_amount_in_cents": 0,
        "donations_raised_amount_in_cents": 0,
        "donations_raised_amount_this_cycle_in_cents": 0,
        "donations_raised_count": 0,
        "donations_raised_count_this_cycle": 0,
        "donations_to_raise_amount_in_cents": 0,
        "email1": "foo@example.com",
        "email1_is_bad": False,
        "email2": None,
        "email2_is_bad": False,
        "email3": None,
        "email3_is_bad": False,
        "email4": None,
        "email4_is_bad": False,
        "emails": [
            {
                "email_address": "foo@example.com",
                "email_number": 1,
                "is_bad": False,
                "is_primary": True,
            }
        ],
        "ethnicity": None,
        "facebook_address": None,
        "facebook_profile_url": None,
        "facebook_updated_at": None,
        "facebook_username": None,
        "fax_number": None,
        "federal_donotcall": False,
        "first_donated_at": None,
        "first_fundraised_at": None,
        "first_invoice_at": None,
        "first_prospect_at": None,
        "first_recruited_at": None,
        "first_supporter_at": "2023-06-22T08:21:00-07:00",
        "first_volunteer_at": None,
        "full_name": "Foo Bar",
        "home_address": None,
        "import_id": None,
        "inferred_party": None,
        "inferred_support_level": None,
        "invoice_payments_amount_in_cents": None,
        "invoice_payments_referred_amount_in_cents": None,
        "invoices_amount_in_cents": None,
        "invoices_count": None,
        "is_absentee_voter": None,
        "is_active_voter": None,
        "is_deceased": False,
        "is_donor": False,
        "is_dropped_from_file": None,
        "is_early_voter": None,
        "is_fundraiser": False,
        "is_ignore_donation_limits": False,
        "is_leaderboardable": True,
        "is_mobile_bad": False,
        "is_permanent_absentee_voter": None,
        "is_possible_duplicate": False,
        "is_profile_private": False,
        "is_profile_searchable": True,
        "is_prospect": False,
        "is_supporter": True,
        "is_survey_question_private": False,
        "language": None,
        "last_call_id": None,
        "last_contacted_at": None,
        "last_contacted_by": None,
        "last_donated_at": None,
        "last_fundraised_at": None,
        "last_invoice_at": None,
        "last_rule_violation_at": None,
        "legal_name": None,
        "locale": None,
        "mailing_address": None,
        "marital_status": None,
        "media_market_name": None,
        "meetup_id": None,
        "meetup_address": None,
        "mobile_normalized": None,
        "nbec_precinct_code": None,
        "nbec_precinct": None,
        "note_updated_at": None,
        "outstanding_invoices_amount_in_cents": None,
        "outstanding_invoices_count": None,
        "overdue_invoices_count": None,
        "page_slug": None,
        "parent": None,
        "parent_id": None,
        "party_member": None,
        "phone_normalized": None,
        "phone_time": None,
        "precinct_code": None,
        "precinct_name": None,
        "prefix": None,
        "previous_party": None,
        "primary_email_id": 1,
        "priority_level": None,
        "priority_level_changed_at": None,
        "profile_content": None,
        "profile_content_html": None,
        "profile_headline": None,
        "received_capital_amount_in_cents": 500,
        "recruiter": None,
        "recruits_count": 0,
        "registered_address": None,
        "registered_at": None,
        "religion": None,
        "rule_violations_count": 0,
        "signup_sources": [],
        "spent_capital_amount_in_cents": 0,
        "submitted_address": None,
        "subnations": [],
        "suffix": None,
        "support_level_changed_at": None,
        "support_probability_score": None,
        "township": None,
        "turnout_probability_score": None,
        "twitter_address": None,
        "twitter_description": None,
        "twitter_followers_count": None,
        "twitter_friends_count": None,
        "twitter_location": None,
        "twitter_login": None,
        "twitter_updated_at": None,
        "twitter_website": None,
        "unsubscribed_at": None,
        "user_submitted_address": None,
        "username": None,
        "voter_updated_at": None,
        "warnings_count": 0,
        "website": None,
        "work_address": None,
    },
    "precinct": None,
}



class Contacts:
    """A class to access the contacts PDI API endpoint."""

    def __init__(self):
        self.url_contacts = self.base_url + "/contacts"
        super().__init__()

    def get_contacts(
        self,
        email: str = None,
        phone: str = None,
        first_name: str = None,
        last_name: str = None,
        zip_code: str = None,
        search_by_email: bool = False,
        limit: int = None,
    ):
        """
        Get a list of Contacts.
        `Args:`
            email: str, email address
            phone: str, phone number
            first_name: str, first name
            last_name: str, last name
            zip code: str, zip code
            search_by_email: bool, whether to search using email address
            limit: int
                The number of contacts to return.
        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        params = {
            "email": email,
            "phone": phone,
            "firstName": first_name,
            "lastName": last_name,
            "zipCode": zip_code,
            "searchByEmail": search_by_email,
        }
        return self._request(self.url_contacts, args=params, limit=limit)

    def create_contact(
        self,
        name_prefix="",
        first_name="",
        last_name="",
        middle_name="",
        name_suffix="",
        nickname="",
        occupation="",
        employer="",
        volunteer_status="",
        donor_status="",
        member_status="",
        date_of_birth=None,
        gender=None,
        pdi_id=None,
    ):
        """
        Create new contact
        `Args:`
            pdiId (string, optional): The pdi identifier. pdiId field is ignored when updating. ,
            namePrefix (string): The name prefix.
            firstName (string): The first name.
            middleName (string): The middle name.
            lastName (string): The last name.
            nameSuffix (string): The name suffix.
            nickname (string): The nickname.
            occupation (string): The occupation.
            employer (string): The employer.
            volunteerStatus (string): The volunteer status.
            Options are: "Prospect", "Active", "Inactive", "None", "" ,
            donorStatus (string): The donor status.
            Options are: "Prospect", "Active", "Inactive", "None", "" ,
            memberStatus (string): The member status.
            Options are: "Prospect", "Active", "Inactive", "None", "" ,
            dateOfBirth (string, optional): The date of birth.
            Format allowed: yyyy-MM-dd ,
            gender (string, optional): The gender.
            Options are: "F", "M", "U"

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        payload = {
            "namePrefix": name_prefix,
            "firstName": first_name,
            "lastName": last_name,
            "nameSuffix": name_suffix,
            "nickname": nickname,
            "middleName": middle_name,
            "occupation": occupation,
            "employer": employer,
            "volunteerStatus": volunteer_status,
            "donorStatus": donor_status,
            "memberStatus": member_status,
            "dateOfBirth": date_of_birth,
            "gender": gender,
            "pdiId": pdi_id,
        }
        return self._request(self.url_contacts, req_type="POST", post_data=payload)

    def get_contact(self, id: str):
        """
        Get a Contact by id.

        `Args:`
            id: str
                The Contact id
        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        # todo not working quite right
        return self._request(f"{self.url_contacts}/{id}")

    def update_contact(
        self,
        id,
        first_name,
        last_name,
        name_prefix="",
        middle_name="",
        name_suffix="",
        nickname="",
        occupation="",
        employer="",
        volunteer_status="",
        donor_status="",
        member_status="",
        date_of_birth=None,
        gender="U",
    ):
        """
        Update Contact
        `Args:`
            namePrefix (string): The name prefix.
            firstName (string): The first name.
            middleName (string): The middle name.
            lastName (string): The last name.
            nameSuffix (string): The name suffix.
            nickname (string): The nickname.
            occupation (string): The occupation.
            employer (string): The employer.
            volunteerStatus (string): The volunteer status.
            Options are: "Prospect", "Active", "Inactive", "None", "" ,
            donorStatus (string): The donor status.
            Options are: "Prospect", "Active", "Inactive", "None", "" ,
            memberStatus (string): The member status.
            Options are: "Prospect", "Active", "Inactive", "None", "" ,
            dateOfBirth (string, optional): The date of birth.
            Format allowed: yyyy-MM-dd ,
            gender (string, optional): The gender.
            Options are: "F", "M", "U"

        `Returns:`
            parsons.Table
                A Parsons table of all the data.
        """
        payload = {
            "namePrefix": name_prefix,
            "firstName": first_name,
            "middleName": middle_name,
            "lastName": last_name,
            "nameSuffix": name_suffix,
            "nickname": nickname,
            "occupation": occupation,
            "employer": employer,
            "volunteerStatus": volunteer_status,
            "donorStatus": donor_status,
            "memberStatus": member_status,
            "dateOfBirth": date_of_birth,
            "gender": gender,
        }
        res = self._request(
            f"{self.url_contacts}/{id}", req_type="PUT", post_data=payload
        )
        if res["code"] == 201:
            return True

    def add_phone(
        self,
        contact_id: int,
        phone_number: str,
        phone_type="Mobile",
        primary=True,
        extension=None,
    ):
        """Add a phone number to a contact
        `Args:`
            contact_id: int
                Unique ID of the contact you'd like to apply the phone_number to
            phone_number: str
            phone_type: str
                Options are `Home`, `Work`, `Direct`, `Mobile`, `Fax`, and `Other. Defaults to
                `Mobile`
            primary: bool
                True indicates that this phone number is the contact's primary phone number
            extension: str
        `Returns:`
            dict
                Response from PDI
        """

        payload = {
            "phoneNumber": phone_number,
            "phoneType": phone_type,
            "isPrimary": primary,
        }

        if extension:
            payload["extension"] = extension

        response = self._request(
            self.url_contacts + f"/{str(contact_id)}/phones",
            req_type="POST",
            post_data=payload,
        )

        return response

    def add_email(self, contact_id: int, email: str, primary=True):
        """Add an email address to a contact
        `Args:`
            contact_id: int
                Unique ID of the contact you'd like to apply the email to
            email: str
            primary: bool
                True indicates that this email address is the contact's primary email
        `Returns:`
            dict
                Response from PDI
        """

        payload = {"emailAddress": email, "isPrimary": primary}

        response = self._request(
            self.url_contacts + f"/{str(contact_id)}/emails",
            req_type="POST",
            post_data=payload,
        )

        return response

    def delete_contact(self, id: str):
        """
        Delete a Question by id.
        `Args:`
            id: str
                The Question id
        """
        return self._request(f"{self.url_contacts}/{id}", req_type="DELETE")



import unittest
import os
import requests_mock
from parsons import VAN
from test.test_van.responses_printed_lists import list_json, single_list_json


class TestSavedLists(unittest.TestCase):
    def setUp(self):

        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters", raise_for_status=False)

    @requests_mock.Mocker()
    def test_get_printed_lists(self, m):

        m.get(self.van.connection.uri + "printedLists", json=list_json)

        result = self.van.get_printed_lists(folder_name="Covington Canvass Turfs")

        self.assertEqual(result.num_rows, 14)

    @requests_mock.Mocker()
    def test_get_printed_list(self, m):

        m.get(self.van.connection.uri + "printedLists/43-0000", json=single_list_json)

        result = self.van.get_printed_list(printed_list_number="43-0000")

        self.assertEqual(result["number"], "43-0000")



import unittest
import os
import requests_mock
from parsons import VAN
from test.utils import validate_list

os.environ["VAN_API_KEY"] = "SOME_KEY"


class TestActivistCodes(unittest.TestCase):
    def setUp(self):

        self.van = VAN(os.environ["VAN_API_KEY"], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_activist_codes(self, m):

        # Create response
        json = {
            "count": 43,
            "items": [
                {
                    "status": "Active",
                    "scriptQuestion": None,
                    "name": "TEST CODE",
                    "mediumName": "TEST CODE",
                    "activistCodeId": 4388538,
                    "shortName": "TC",
                    "type": "Action",
                    "description": None,
                }
            ],
            "nextPageLink": None,
        }

        m.get(self.van.connection.uri + "activistCodes", json=json)

        # Expected Structure
        expected = [
            "status",
            "scriptQuestion",
            "name",
            "mediumName",
            "activistCodeId",
            "shortName",
            "type",
            "description",
        ]

        # Assert response is expected structure
        self.assertTrue(validate_list(expected, self.van.get_activist_codes()))

        # To Do: Test what happens when it doesn't find any ACs

    @requests_mock.Mocker()
    def test_get_activist_code(self, m):

        # Create response
        json = {
            "status": "Active",
            "scriptQuestion": "null",
            "name": "Anti-Choice",
            "mediumName": "Anti",
            "activistCodeId": 4135099,
            "shortName": "AC",
            "type": "Constituency",
            "description": "A person who has been flagged as anti-choice.",
        }

        m.get(self.van.connection.uri + "activistCodes/4388538", json=json)

        self.assertEqual(json, self.van.get_activist_code(4388538))

    @requests_mock.Mocker()
    def test_toggle_activist_code(self, m):

        # Test apply activist code
        m.post(
            self.van.connection.uri + "people/2335282/canvassResponses", status_code=204
        )
        self.assertTrue(self.van.toggle_activist_code(2335282, 4429154, "apply"), 204)

        # Test remove activist code
        m.post(
            self.van.connection.uri + "people/2335282/canvassResponses", status_code=204
        )
        self.assertTrue(self.van.toggle_activist_code(2335282, 4429154, "remove"), 204)

    @requests_mock.Mocker()
    def test_apply_activist_code(self, m):

        # Test apply activist code
        m.post(
            self.van.connection.uri + "people/2335282/canvassResponses", status_code=204
        )
        self.assertEqual(self.van.apply_activist_code(2335282, 4429154), 204)

    @requests_mock.Mocker()
    def test_remove_activist_code(self, m):

        # Test remove activist code
        m.post(
            self.van.connection.uri + "people/2335282/canvassResponses", status_code=204
        )
        self.assertEqual(self.van.remove_activist_code(2335282, 4429154), 204)



import json
import logging
import requests
import time

from parsons.etl.table import Table
from parsons.utilities.check_env import check

logger = logging.getLogger(__name__)


class RedashTimeout(Exception):
    pass


class RedashQueryFailed(Exception):
    pass


class Redash(object):
    """
    Instantiate Redash Class

    `Args:`
        base_url: str
            The base url for your redash instance (excluding the final /)
        user_api_key: str
            The user API key found in the User's profile screen
        pause_time int
            Specify time between polling for refreshed queries (Defaults to 3 seconds)
        verify: bool
            For https requests, should the certificate be verified (Defaults to True)
    `Returns:`
        Redash Class
    """

    def __init__(
        self,
        base_url=None,
        user_api_key=None,
        pause_time=3,
        timeout=0,  # never timeout
        verify=True,
    ):
        self.base_url = check("REDASH_BASE_URL", base_url)
        self.user_api_key = check("REDASH_USER_API_KEY", user_api_key, optional=True)
        self.pause = int(check("REDASH_PAUSE_TIME", pause_time, optional=True))
        self.timeout = int(check("REDASH_TIMEOUT", timeout, optional=True))

        self.verify = verify  # for https requests
        self.session = requests.Session()
        if user_api_key:
            self.session.headers.update({"Authorization": f"Key {user_api_key}"})

    def _catch_runtime_error(self, res):
        if res.status_code != 200:
            raise RuntimeError(
                f"Error. Status code: {res.status_code}. Reason: {res.reason}"
            )

    def _poll_job(self, session, job, query_id):
        start_secs = time.time()
        while job["status"] not in (3, 4):
            if self.timeout and start_secs + self.timeout < time.time():
                raise RedashTimeout(f"Redash timeout: {self.timeout}")
            poll_url = "{}/api/jobs/{}".format(self.base_url, job["id"])
            response = session.get(poll_url, verify=self.verify)
            response_json = response.json()
            job = response_json.get(
                "job",
                {
                    "status": "Error NO JOB IN RESPONSE: {}".format(
                        json.dumps(response_json)
                    )
                },
            )
            logger.debug(
                "poll url:%s id:%s status:%s err:%s",
                poll_url,
                query_id,
                job["status"],
                job.get("error"),
            )
            time.sleep(self.pause)

        if job["status"] == 3:  # 3 = completed
            return job["query_result_id"]
        elif job["status"] == 4:  # 3 = ERROR
            raise RedashQueryFailed(
                "Redash Query {} failed: {}".format(query_id, job["error"])
            )

    def get_data_source(self, data_source_id):
        """
        Get a data source.

        `Args:`
            data_source_id: int or str
                ID of data source.
        `Returns`:
            Data source json object
        """
        res = self.session.get(f"{self.base_url}/api/data_sources/{data_source_id}")
        self._catch_runtime_error(res)
        return res.json()

    def update_data_source(
        self, data_source_id, name, type, dbName, host, password, port, user
    ):
        """
        Update a data source.

        `Args:`
            data_source_id: str or int
                ID of data source.
            name: str
                Name of data source.
            type: str
                Type of data source.
            dbname: str
                Database name of data source.
            host: str
                Host of data source.
            password: str
                Password of data source.
            port: int or str
                Port of data source.
            user: str
                Username of data source.
        `Returns:`
            ``None``
        """
        self._catch_runtime_error(
            self.session.post(
                f"{self.base_url}/api/data_sources/{data_source_id}",
                json={
                    "name": name,
                    "type": type,
                    "options": {
                        "dbname": dbName,
                        "host": host,
                        "password": password,
                        "port": port,
                        "user": user,
                    },
                },
            )
        )

    def get_fresh_query_results(self, query_id=None, params=None):
        """
        Make a fresh query result and get back the CSV http response object back
        with the CSV string in result.content

        `Args:`
            query_id: str or int
                The query id of the query
            params: dict
                If there are values for the redash query parameters
                (described https://redash.io/help/user-guide/querying/query-parameters
                e.g. "{{datelimit}}" in the query),
                then this is a dict that will pass the parameters in the POST.
                We add the "p_" prefix for parameters, so if your query had ?p_datelimit=....
                in the url, you should just set 'datelimit' in params here.
                If you set this with REDASH_QUERY_PARAMS environment variable instead of passing
                the values, then you must include the "p_" prefixes and it should be a single
                url-encoded string as you would see it in the URL bar.
        `Returns:`
            Table Class
        """
        query_id = check("REDASH_QUERY_ID", query_id, optional=True)
        params_from_env = check("REDASH_QUERY_PARAMS", "", optional=True)
        redash_params = (
            {"p_%s" % k: str(v).replace("'", "''") for k, v in params.items()}
            if params
            else {}
        )

        response = self.session.post(
            f"{self.base_url}/api/queries/{query_id}/refresh?{params_from_env}",
            params=redash_params,
            verify=self.verify,
        )

        if response.status_code != 200:
            raise RedashQueryFailed(
                f"Refresh failed for query {query_id}. {response.text}"
            )

        job = response.json()["job"]
        result_id = self._poll_job(self.session, job, query_id)
        if result_id:
            response = self.session.get(
                f"{self.base_url}/api/queries/{query_id}/results/{result_id}.csv",
                verify=self.verify,
            )
            if response.status_code != 200:
                raise RedashQueryFailed(
                    f"Failed getting results for query {query_id}. {response.text}"
                )
        else:
            raise RedashQueryFailed(
                f"Failed getting result {query_id}. {response.text}"
            )
        return Table.from_csv_string(response.text)

    def get_cached_query_results(self, query_id=None, query_api_key=None):
        """
        Get the results from a cached query result and get back the CSV http response object back
        with the CSV string in result.content

        `Args:`
            query_id: str or int
                The query id of the query
            query_api_key: str
                If you did not supply a user_api_key on the Redash object, then you can
                supply a query_api_key to get cached results back anonymously.
        `Returns:`
            Table Class
        """
        query_id = check("REDASH_QUERY_ID", query_id)
        query_api_key = check("REDASH_QUERY_API_KEY", query_api_key, optional=True)
        params = {}
        if not self.user_api_key and query_api_key:
            params["api_key"] = query_api_key
        response = self.session.get(
            f"{self.base_url}/api/queries/{query_id}/results.csv",
            params=params,
            verify=self.verify,
        )
        if response.status_code != 200:
            raise RedashQueryFailed(
                f"Failed getting results for query {query_id}. {response.text}"
            )
        return Table.from_csv_string(response.text)

    @classmethod
    def load_to_table(cls, refresh=True, **kwargs):
        """
        Fast classmethod makes the appropriate query type (refresh or cached)
        based on which arguments are supplied.

        `Args:`
            base_url: str
                The base url for your redash instance (excluding the final /)
            query_id: str or int
                The query id of the query
            user_api_key: str
                The user API key found in the User's profile screen required for refresh queries
            query_api_key: str
                If you did not supply a user_api_key on the Redash object, then you can
                supply a query_api_key to get cached results back anonymously.
            pause_time int
                Specify time between polling for refreshed queries (Defaults to 3 seconds)
            verify: bool
                For https requests, should the certificate be verified (Defaults to True)
            refresh: bool
                Refresh results or cached. (Defaults to True unless a query_api_key IS supplied)
            params: dict
                For refresh queries, if there are parameters in the query,
                then this is a dict that will pass the parameters in the POST.
                We add the "p_" prefix for parameters, so if your query had ?p_datelimit=....
                in the url, you should just set 'datelimit' in params here.

        `Returns:`
            Table Class
        """
        initargs = {
            a: kwargs.get(a)
            for a in ("base_url", "user_api_key", "pause_time", "timeout", "verify")
            if a in kwargs
        }
        obj = cls(**initargs)
        if not refresh or kwargs.get("query_api_key"):
            return obj.get_cached_query_results(
                kwargs.get("query_id"), kwargs.get("query_api_key")
            )
        else:
            return obj.get_fresh_query_results(
                kwargs.get("query_id"), kwargs.get("params")
            )



from test.utils import mark_live_test

from parsons import Table

from contextlib import contextmanager
from requests.exceptions import HTTPError

# import json
import pytest

#
# Fixtures and constants
#

# Set these to be valid for a qa account

# get_flag_ids
QA_NUM_FLAG_IDS = 34

# get_flag_id
QA_REAL_FLAG_ID = "RtK8WQc4uTnuJYGCHMDpmA=="
QA_INVALID_FLAG_ID = "jE0muNuPKfSTg8hTen10kA=="
QA_MALFORMED_FLAG_ID = QA_INVALID_FLAG_ID[:-1]

# create a mark that expects a failure from HTTPError
xfail_http_error = pytest.mark.xfail(raises=HTTPError, strict=True)


@pytest.fixture
def cleanup_flag_id():
    def delete_flag_id(pdi, flag_id):
        pdi.delete_flag_id(flag_id)

    yield delete_flag_id


@pytest.fixture
def create_temp_flag_id():
    @contextmanager
    def temp_flag_id(pdi, my_flag_id=None):
        flag_id = my_flag_id or pdi.create_flag_id("amm", True)
        print(flag_id)

        yield flag_id

        if not my_flag_id:
            pdi.delete_flag_id(flag_id)

    yield temp_flag_id


#
# Tests
#


@mark_live_test
@pytest.mark.parametrize("limit", [None, 5, 15])
def test_get_flag_ids(live_pdi, limit):
    flag_ids = live_pdi.get_flag_ids(limit=limit)

    expected_columns = ["id", "flagId", "flagIdDescription", "compile", "isDefault"]
    expected_num_rows = limit or QA_NUM_FLAG_IDS

    assert isinstance(flag_ids, Table)
    assert flag_ids.columns == expected_columns
    assert flag_ids.num_rows == expected_num_rows


@mark_live_test
@pytest.mark.parametrize(
    "id",
    [
        pytest.param(QA_REAL_FLAG_ID),
        pytest.param(QA_INVALID_FLAG_ID, marks=[xfail_http_error]),
    ],
)
def test_get_flag_id(live_pdi, id):
    flag_id = live_pdi.get_flag_id(id)

    expected_keys = ["id", "flagId", "flagIdDescription", "compile", "isDefault"]

    assert isinstance(flag_id, dict)
    assert list(flag_id.keys()) == expected_keys


@mark_live_test
@pytest.mark.parametrize(
    ["flag_id", "is_default"],
    [
        pytest.param(None, True, marks=[xfail_http_error]),
        pytest.param("amm", None, marks=[xfail_http_error]),
        pytest.param("amm", True),
    ],
)
def test_create_flag_id(live_pdi, cleanup_flag_id, flag_id, is_default):
    flag_id = live_pdi.create_flag_id(flag_id, is_default)

    cleanup_flag_id(live_pdi, flag_id)


@mark_live_test
@pytest.mark.parametrize(
    ["my_flag_id"],
    [
        pytest.param(None),
        pytest.param(QA_INVALID_FLAG_ID),
        pytest.param(QA_MALFORMED_FLAG_ID, marks=[xfail_http_error]),
    ],
)
def test_delete_flag_id(live_pdi, create_temp_flag_id, my_flag_id):
    with create_temp_flag_id(live_pdi, my_flag_id) as flag_id:
        did_delete = live_pdi.delete_flag_id(flag_id)

    assert did_delete


@mark_live_test
@pytest.mark.parametrize(
    ["my_flag_id"],
    [
        pytest.param(None),
        pytest.param(QA_INVALID_FLAG_ID, marks=[xfail_http_error]),
        pytest.param(QA_MALFORMED_FLAG_ID, marks=[xfail_http_error]),
    ],
)
def test_update_flag_id(live_pdi, create_temp_flag_id, my_flag_id):
    with create_temp_flag_id(live_pdi, my_flag_id) as flag_id:
        # flag initial state:
        # {"id":flag_id,"flagId":"amm","flagIdDescription":null,"compile":"","isDefault":false}  # noqa
        id = live_pdi.update_flag_id(flag_id, "bnh", True)
        assert id == flag_id

        expected_dict = {
            "id": flag_id,
            "flagId": "bnh",
            "flagIdDescription": None,
            "compile": "",
            "isDefault": True,
        }

        flag_id_dict = live_pdi.get_flag_id(flag_id)

        assert expected_dict == flag_id_dict
